# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import contextlib
import time


@contextlib.contextmanager
def log_timer_debug(msg, logger):
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        logger.debug(msg + f" Completed in {end_time-start_time} seconds")


@contextlib.contextmanager
def log_timer_info(msg, logger):
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        logger.info(msg + f" Completed in {end_time-start_time} seconds")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from decouple import config

PYTHON_LOG_LEVEL = config("PYTHON_LOG_LEVEL", "INFO")

NO_REDIS = config("NO_REDIS", False, cast=bool)
DISABLE_TAAR_LITE = config("DISABLE_TAAR_LITE", False, cast=bool)
DISABLE_ENSEMBLE = config("DISABLE_ENSEMBLE", False, cast=bool)

REDIS_HOST = config("REDIS_HOST", "localhost", cast=str)
REDIS_PORT = config("REDIS_PORT", 6379, cast=int)

# These are configurations that are specific to the TAAR library
TAAR_MAX_RESULTS = config("TAAR_MAX_RESULTS", default=10, cast=int)

TAARLITE_MAX_RESULTS = config("TAARLITE_MAX_RESULTS", default=4, cast=int)

STATSD_HOST = config("STATSD_HOST", default="localhost", cast=str)
STATSD_PORT = config("STATSD_PORT", default=8125, cast=int)

TAAR_ENSEMBLE_BUCKET = config("TAAR_ENSEMBLE_BUCKET", default="test_ensemble_bucket")
TAAR_ENSEMBLE_KEY = config("TAAR_ENSEMBLE_KEY", default="test_ensemble_key")

TAAR_WHITELIST_BUCKET = config("TAAR_WHITELIST_BUCKET", default="test_whitelist_bucket")
TAAR_WHITELIST_KEY = config("TAAR_WHITELIST_KEY", default="test_whitelist_key")

TAAR_ITEM_MATRIX_BUCKET = config(
    "TAAR_ITEM_MATRIX_BUCKET", default="test_matrix_bucket"
)
TAAR_ITEM_MATRIX_KEY = config("TAAR_ITEM_MATRIX_KEY", default="test_matrix_key")
TAAR_ADDON_MAPPING_BUCKET = config(
    "TAAR_ADDON_MAPPING_BUCKET", default="test_mapping_bucket"
)
TAAR_ADDON_MAPPING_KEY = config("TAAR_ADDON_MAPPING_KEY", default="test_mapping_key")

TAAR_LOCALE_BUCKET = config("TAAR_LOCALE_BUCKET", default="test_locale_bucket")
TAAR_LOCALE_KEY = config("TAAR_LOCALE_KEY", default="test_locale_key")


TAAR_SIMILARITY_BUCKET = config(
    "TAAR_SIMILARITY_BUCKET", default="test_similarity_bucket"
)
TAAR_SIMILARITY_DONOR_KEY = config(
    "TAAR_SIMILARITY_DONOR_KEY", default="test_similarity_donor_key"
)
TAAR_SIMILARITY_LRCURVES_KEY = config(
    "TAAR_SIMILARITY_LRCURVES_KEY", default="test_similarity_lrcurves_key"
)


# TAAR-lite configuration below

TAARLITE_GUID_COINSTALL_BUCKET = config(
    "TAARLITE_GUID_COINSTALL_BUCKET", "telemetry-parquet"
)
TAARLITE_GUID_COINSTALL_KEY = config(
    "TAARLITE_GUID_COINSTALL_KEY", "taar/lite/guid_coinstallation.json"
)

TAARLITE_GUID_RANKING_KEY = config(
    "TAARLITE_GUID_RANKING_KEY", "taar/lite/guid_install_ranking.json"
)

# 4 hour liviliness for TAARLITE data
TAARLITE_TTL = config("TAARLITE_TTL", 60 * 60 * 4, cast=int)


# TAARlite needs redis backed mutex's to protect critical sections
TAARLITE_MUTEX_TTL = config("TAARLITE_MUTEX_TTL", 60 * 60, cast=int)


# Set a default TAARLite mutex TTL of 1 hour to fully populate the
# redis cache
TAARLITE_MUTEX_TTL = config("TAARLITE_MUTEX_TTL", 60 * 60, cast=int)

TAARLITE_TRUNCATE = config("TAARLITE_TRUNCATE", TAARLITE_MAX_RESULTS * 5, cast=int)

# Bigtable config
BIGTABLE_PROJECT_ID = config(
    "BIGTABLE_PROJECT_ID", default="cfr-personalization-experiment"
)
BIGTABLE_INSTANCE_ID = config("BIGTABLE_INSTANCE_ID", default="taar-profile")
BIGTABLE_TABLE_ID = config("BIGTABLE_TABLE_ID", default="taar_profile")



"""
This module re-orders the (GUID, weight) 2-tuples using
numpy.random.choice
"""

import numpy as np


def in_experiment(client_id, xp_prob=0.5):
    """
    Return whether or not this client_id is in the experiment.

    xp_prob is a probability between 0.0 and 1.0 which is the
    chance that the experimental branch is selected.
    """
    hex_client = "".join([c for c in client_id.lower() if c in "abcdef0123456789"])
    int_client = int(hex_client, 16)
    return int((int_client % 100) < (xp_prob * 100))


def reorder_guids(guid_weight_tuples, size=None):
    """
    This reorders (GUID, weight) 2-tuples based on the weight using
    random selection, without replacement.

    @size denotes the length of the output.
    """
    if guid_weight_tuples is None or len(guid_weight_tuples) == 0:
        return []

    weights = np.array([weight for (guid, weight) in guid_weight_tuples])
    guids = [guid for (guid, weight) in guid_weight_tuples]
    guid_map = dict(zip(guids, guid_weight_tuples))

    if size is None:
        size = len(guids)
    else:
        size = min(size, len(guids))

    # Normalize the weights so that they're probabilities
    # Scale first, weights can be negative (for example, collaborative filtering similarity scores)
    scaled_weights = weights - np.min(weights) + np.finfo(float).eps
    probabilities = scaled_weights / np.sum(scaled_weights)

    choices = np.random.choice(guids, size=size, replace=False, p=probabilities)
    return [guid_map[guid] for guid in choices]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.recommenders.base_recommender import AbstractRecommender
from itertools import groupby
from scipy.spatial import distance
from taar.logs.interfaces import IMozLogging, ITAARCache
import numpy as np

FLOOR_DISTANCE_ADJUSTMENT = 0.001

CATEGORICAL_FEATURES = ["geo_city", "locale", "os"]
CONTINUOUS_FEATURES = [
    "subsession_length",
    "bookmark_count",
    "tab_open_count",
    "total_uri",
    "unique_tlds",
]


class SimilarityRecommender(AbstractRecommender):
    """ A recommender class that returns top N addons based on the
    client similarity with a set of candidate addon donors.

    Several telemetry fields are used to compute pairwise similarity
    with the donors and similarities are converted into a likelihood
    ratio of being a good match versus not being a good match. These
    quantities are then used to rank specific addons for
    recommendation.

    This will load a json file containing updated list of addon donors
    updated periodically by a separate weekly process using
    Longitdudinal Telemetry data.

    This recommender may provide useful recommendations when
    collaborative_recommender may not work.
    """

    def __init__(self, ctx):
        self._ctx = ctx

        self._cache = self._ctx[ITAARCache]

        self.logger = self._ctx[IMozLogging].get_logger("taar")

    def _get_cache(self, extra_data):
        tmp = extra_data.get("cache", None)
        if tmp is None:
            tmp = self._cache.cache_context()
        return tmp

    """
    End private properties
    """

    def can_recommend(self, client_data, extra_data={}):
        cache = self._get_cache(extra_data)
        # We can't recommend if we don't have our data files.
        if cache["donors_pool"] is None or cache["lr_curves"] is None:
            return False

        # Check that the client info contains a non-None value for each required
        # telemetry field.
        REQUIRED_FIELDS = CATEGORICAL_FEATURES + CONTINUOUS_FEATURES

        has_fields = all(
            [client_data.get(f, None) is not None for f in REQUIRED_FIELDS]
        )
        if not has_fields:
            # Can not add extra info because client_id may not be available.
            self.logger.error("Unusable client data encountered")
        return has_fields

    def get_lr(self, score, cache):
        """Compute a :float: likelihood ratio from a provided similarity score when compared
        to two probability density functions which are computed and pre-loaded during init.

        The numerator indicates the probability density that a particular similarity score
        corresponds to a 'good' addon donor, i.e. a client that is similar in the sense of
        telemetry variables. The denominator indicates the probability density that a particular
        similarity score corresponds to a 'poor' addon donor

        :param score: A similarity score between a pair of objects.
        :returns: The approximate float likelihood ratio corresponding to provided score.
        """
        # Find the index of the closest value that was precomputed in lr_curves
        # This will significantly speed up |get_lr|.

        # The lr_curves_cache is a list of scalar distance
        # measurements
        lr_curves_cache = np.array([s[0] for s in cache["lr_curves"]])

        # np.argmin produces the index to the part of the curve
        # where distance is the smallest to the score which we are
        # inspecting currently.
        idx = np.argmin(abs(score - lr_curves_cache))

        numer_val = cache["lr_curves"][idx][1][0]
        denum_val = cache["lr_curves"][idx][1][1]

        # Compute LR based on numerator and denominator values
        return float(numer_val) / float(denum_val)

    # # # CAUTION! # # #
    # Any changes to this function must be reflected in the corresponding ETL job.
    # https://github.com/mozilla/python_mozetl/blob/master/mozetl/taar/taar_similarity.py
    #
    def compute_clients_dist(self, client_data, cache):
        client_categorical_feats = [
            client_data.get(specified_key) for specified_key in CATEGORICAL_FEATURES
        ]
        client_continuous_feats = [
            client_data.get(specified_key) for specified_key in CONTINUOUS_FEATURES
        ]

        # Compute the distances between the user and the cached continuous features.
        cont_features = distance.cdist(
            cache["continuous_features"],
            np.array([client_continuous_feats]),
            "canberra",
        )

        # Compute the distances between the user and the cached categorical features.
        cat_features = np.array(
            [
                [distance.hamming(x, client_categorical_feats)]
                for x in cache["categorical_features"]
            ]
        )

        # See the "Note about cdist optimization" in README.md for why we only use cdist once.

        # Take the product of similarities to attain a univariate similarity score.
        # Note that the addition of 0.001 to the continuous features
        # sets a floor value to the distance in continuous similarity
        # scores.  There is no such floor value set for categorical
        # features so this adjustment prioritizes categorical
        # similarity over continous similarity
        return (cont_features + FLOOR_DISTANCE_ADJUSTMENT) * cat_features

    def get_similar_donors(self, client_data, cache):
        """Computes a set of :float: similarity scores between a client and a set of candidate
        donors for which comparable variables have been measured.

        A custom similarity metric is defined in this function that combines the Hamming distance
        for categorical variables with the Canberra distance for continuous variables into a
        univariate similarity metric between the client and a set of candidate donors loaded during
        init.

        :param client_data: a client data payload including a subset fo telemetry fields.
        :return: the sorted approximate likelihood ratio (np.array) corresponding to the
                 internally computed similarity score and a list of indices that link
                 each LR score with the related donor in the |self.donors_pool|.
        """
        # Compute the distance between self and any comparable client.
        distances = self.compute_clients_dist(client_data, cache)

        # Compute the LR based on precomputed distributions that relate the score
        # to a probability of providing good addon recommendations.

        lrs_from_scores = np.array(
            [self.get_lr(distances[i], cache) for i in range(cache["num_donors"])]
        )

        # Sort the LR values (descending) and return the sorted values together with
        # the original indices.
        indices = (-lrs_from_scores).argsort()
        return lrs_from_scores[indices], indices

    def _recommend(self, client_data, limit, extra_data={}):
        cache = self._get_cache(extra_data)

        donor_set_ranking, indices = self.get_similar_donors(client_data, cache)
        donor_log_lrs = np.log(donor_set_ranking)
        # 1.0 corresponds to a log likelihood ratio of 0 meaning that donors are equally
        # likely to be 'good'. A value > 0.0 is sufficient, but we like this to be high.
        if donor_log_lrs[0] < 0.1:
            self.logger.warning(
                "Addons recommended with very low similarity score, perhaps donor set is unrepresentative",
                extra={"maximum_similarity": donor_set_ranking[0]},
            )

        # Retrieve the indices of the highest ranked donors and then append their
        # installed addons.
        index_lrs_iter = zip(indices[donor_log_lrs > 0.0], donor_log_lrs)
        recommendations = []
        for (index, lrs) in index_lrs_iter:
            for term in cache["donors_pool"][index]["active_addons"]:
                candidate = (term, lrs)
                recommendations.append(candidate)
        # Sort recommendations on key (guid name)
        recommendations = sorted(recommendations, key=lambda x: x[0])
        recommendations_out = []
        # recommendations must be sorted for this to work.
        for guid_key, group in groupby(recommendations, key=lambda x: x[0]):
            recommendations_out.append((guid_key, sum(j for i, j in group)))
        # now re-sort on the basis of LLR.
        recommendations_out = sorted(recommendations_out, key=lambda x: -x[1])

        log_data = (
            client_data["client_id"],
            str([r[0] for r in recommendations_out[:limit]]),
        )
        self.logger.debug(
            "similarity_recommender_triggered, "
            "client_id: [%s], guids: [%s]" % log_data
        )
        return recommendations_out

    def recommend(self, client_data, limit, extra_data={}):
        recommendations_out = self._recommend(client_data, limit, extra_data)
        return recommendations_out[:limit]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from taar.logs.interfaces import ITAARCache
from taar.recommenders.ensemble_recommender import EnsembleRecommender
import mock
import contextlib
import fakeredis
from taar.recommenders.redis_cache import TAARCacheRedis
from .noop_fixtures import (
    noop_taarlocale_dataload,
    noop_taarcollab_dataload,
    noop_taarlite_dataload,
    noop_taarsimilarity_dataload,
)
from .mocks import MockRecommenderFactory

from markus import TIMING
from markus.testing import MetricsMock

EXPECTED = {"collaborative": 1000, "similarity": 100, "locale": 10}


def noop_loaders(stack):
    stack = noop_taarlocale_dataload(stack)
    stack = noop_taarcollab_dataload(stack)
    stack = noop_taarlite_dataload(stack)
    stack = noop_taarsimilarity_dataload(stack)
    return stack


@contextlib.contextmanager
def mock_install_mock_ensemble_data(ctx):
    DATA = {"ensemble_weights": EXPECTED}

    WHITELIST_DATA = [
        "2.0@disconnect.me",
        "@contain-facebook",
        "@testpilot-containers",
        "CookieAutoDelete@kennydo.com",
        "FirefoxColor@mozilla.com",
        "adblockultimate@adblockultimate.net",
        "addon@darkreader.org",
        "adguardadblocker@adguard.com",
        "adnauseam@rednoise.org",
        "clearcache@michel.de.almeida",
        "copyplaintext@eros.man",
        "default-bookmark-folder@gustiaux.com",
        "enhancerforyoutube@maximerf.addons.mozilla.org",
        "extension@one-tab.com",
        "extension@tabliss.io",
        "firefox-addon@myki.co",
        "firefox@ghostery.com",
        "forecastfox@s3_fix_version",
        "forget-me-not@lusito.info",
        "foxyproxy@eric.h.jung",
        "foxytab@eros.man",
        "gmailnoads@mywebber.com",
    ]

    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None
        stack.enter_context(
            mock.patch.object(TAARCacheRedis, "_fetch_ensemble_weights", return_value=DATA, )
        )

        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_whitelist", return_value=WHITELIST_DATA,
            )
        )

        stack = noop_loaders(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


def test_weight_cache(test_ctx):
    with mock_install_mock_ensemble_data(test_ctx):
        factory = MockRecommenderFactory()
        test_ctx["recommender_factory"] = factory

        test_ctx["recommender_map"] = {
            "collaborative": factory.create("collaborative"),
            "similarity": factory.create("similarity"),
            "locale": factory.create("locale"),
        }

        r = EnsembleRecommender(test_ctx)
        actual = r.getWeights()
        assert EXPECTED == actual


def test_recommendations(test_ctx):
    with MetricsMock() as mm:
        with mock_install_mock_ensemble_data(test_ctx):
            EXPECTED_RESULTS = [
                ("ghi", 3430.0),
                ("def", 3320.0),
                ("ijk", 3200.0),
                ("hij", 3100.0),
                ("lmn", 420.0),
            ]

            factory = MockRecommenderFactory()
            test_ctx["recommender_factory"] = factory

            test_ctx["recommender_map"] = {
                "collaborative": factory.create("collaborative"),
                "similarity": factory.create("similarity"),
                "locale": factory.create("locale"),
            }
            r = EnsembleRecommender(test_ctx)
            client = {"client_id": "12345"}  # Anything will work here

            recommendation_list = r.recommend(client, 5)
            assert isinstance(recommendation_list, list)
            assert recommendation_list == EXPECTED_RESULTS

            assert mm.has_record(TIMING, "taar.ensemble_recommend")
            assert mm.has_record(TIMING, "taar.collaborative_recommend")
            assert mm.has_record(TIMING, "taar.locale_recommend")
            assert mm.has_record(TIMING, "taar.similarity_recommend")


def test_preinstalled_guids(test_ctx):
    with mock_install_mock_ensemble_data(test_ctx):
        EXPECTED_RESULTS = [
            ("ghi", 3430.0),
            ("ijk", 3200.0),
            ("lmn", 420.0),
            ("klm", 409.99999999999994),
            ("abc", 23.0),
        ]

        factory = MockRecommenderFactory()
        test_ctx["recommender_factory"] = factory

        test_ctx["recommender_map"] = {
            "collaborative": factory.create("collaborative"),
            "similarity": factory.create("similarity"),
            "locale": factory.create("locale"),
        }
        r = EnsembleRecommender(test_ctx)

        # 'hij' should be excluded from the suggestions list
        # The other two addon GUIDs 'def' and 'jkl' will never be
        # recommended anyway and should have no impact on results
        client = {"client_id": "12345", "installed_addons": ["def", "hij", "jkl"]}

        recommendation_list = r.recommend(client, 5)
        print(recommendation_list)
        assert isinstance(recommendation_list, list)
        assert recommendation_list == EXPECTED_RESULTS


def test_mock_client_ids(test_ctx):
    with mock_install_mock_ensemble_data(test_ctx):

        EXPECTED_RESULTS = [
            ("2.0@disconnect.me", 0.17),
            ("@contain-facebook", 0.25),
            ("@testpilot-containers", 0.72),
            ("CookieAutoDelete@kennydo.com", 0.37),
            ("FirefoxColor@mozilla.com", 0.32),
        ]

        factory = MockRecommenderFactory()
        test_ctx["recommender_factory"] = factory

        test_ctx["recommender_map"] = {
            "collaborative": factory.create("collaborative"),
            "similarity": factory.create("similarity"),
            "locale": factory.create("locale"),
        }
        r = EnsembleRecommender(test_ctx)

        # 'hij' should be excluded from the suggestions list
        # The other two addon GUIDs 'def' and 'jkl' will never be
        # recommended anyway and should have no impact on results
        client = {"client_id": "11111"}

        recommendation_list = r.recommend(client, 5)
        assert isinstance(recommendation_list, list)
        assert recommendation_list == EXPECTED_RESULTS



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.recommenders.recommendation_manager import RecommendationManager
from taar.recommenders.base_recommender import AbstractRecommender

from .noop_fixtures import (
    noop_taarlocale_dataload,
    noop_taarcollab_dataload,
    noop_taarsimilarity_dataload,
    noop_taarlite_dataload,
)

from .mocks import MockRecommenderFactory

import operator
from functools import reduce

import numpy as np
from markus import TIMING
from markus.testing import MetricsMock

import mock
import contextlib
import fakeredis
from taar.recommenders.redis_cache import TAARCacheRedis


@contextlib.contextmanager
def mock_install_mock_curated_data(ctx):
    mock_data = []
    for i in range(20):
        mock_data.append(str(i) * 16)

    mock_ensemble_weights = {
        "ensemble_weights": {"collaborative": 1000, "similarity": 100, "locale": 10,}
    }

    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None

        stack.enter_context(
            mock.patch.object(TAARCacheRedis, "_fetch_whitelist", return_value=mock_data)
        )
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "_fetch_ensemble_weights",
                return_value=mock_ensemble_weights,
            )
        )

        stack = noop_taarlite_dataload(stack)
        stack = noop_taarcollab_dataload(stack)
        stack = noop_taarlocale_dataload(stack)
        stack = noop_taarsimilarity_dataload(stack)

        stack.enter_context(
            mock.patch.object(TAARCacheRedis, "_fetch_whitelist", return_value=mock_data)
        )

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        class DefaultMockProfileFetcher:
            def get(self, client_id):
                return {"client_id": client_id}

        mock_fetcher = DefaultMockProfileFetcher()

        ctx["profile_fetcher"] = mock_fetcher
        ctx["recommender_factory"] = MockRecommenderFactory()

        # Initialize redis
        TAARCacheRedis.get_instance(ctx).safe_load_data()

        yield stack


class StubRecommender(AbstractRecommender):
    """ A shared, stub recommender that can be used for testing.
    """

    def __init__(self, can_recommend, stub_recommendations):
        self._can_recommend = can_recommend
        self._recommendations = stub_recommendations

    def can_recommend(self, client_info, extra_data={}):
        return self._can_recommend

    def recommend(self, client_data, limit, extra_data={}):
        return self._recommendations


def test_none_profile_returns_empty_list(test_ctx):
    with mock_install_mock_curated_data(test_ctx):

        class MockProfileFetcher:
            def get(self, client_id):
                return None

        test_ctx["profile_fetcher"] = MockProfileFetcher()

        rec_manager = RecommendationManager(test_ctx)
        assert rec_manager.recommend("random-client-id", 10) == []


def test_simple_recommendation(test_ctx):
    # Fix the random seed so that we get stable results between test
    # runs
    np.random.seed(seed=42)

    with mock_install_mock_curated_data(test_ctx):
        EXPECTED_RESULTS = [
            ("def", 3320.0),
            ("klm", 409.99999999999994),
            ("hij", 3100.0),
            ("ijk", 3200.0),
            ("ghi", 3430.0),
            ("lmn", 420.0),
            ("jkl", 400.0),
            ("abc", 23.0),
            ("fgh", 22.0),
            ("efg", 21.0)
        ]

        with MetricsMock() as mm:
            manager = RecommendationManager(test_ctx)
            recommendation_list = manager.recommend("some_ignored_id", 10)

            assert isinstance(recommendation_list, list)
            assert recommendation_list == EXPECTED_RESULTS

            assert mm.has_record(TIMING, stat="taar.profile_recommendation")


def test_fixed_client_id_valid(test_ctx):
    with mock_install_mock_curated_data(test_ctx):
        manager = RecommendationManager(test_ctx)
        recommendation_list = manager.recommend("111111", 10)
        assert len(recommendation_list) == 10


def test_fixed_client_id_empty_list(test_ctx):
    class NoClientFetcher:
        def get(self, client_id):
            return None

    with mock_install_mock_curated_data(test_ctx):
        test_ctx["profile_fetcher"] = NoClientFetcher()

        manager = RecommendationManager(test_ctx)
        recommendation_list = manager.recommend("not_a_real_client_id", 10)

        assert len(recommendation_list) == 0


def test_experimental_randomization(test_ctx):
    with mock_install_mock_curated_data(test_ctx):

        manager = RecommendationManager(test_ctx)
        raw_list = manager.recommend("111111", 10)

        # Clobber the experiment probability to be 100% to force a
        # reordering.
        test_ctx["TAAR_EXPERIMENT_PROB"] = 1.0

        manager = RecommendationManager(test_ctx)
        rand_list = manager.recommend("111111", 10)

        """
        The two lists should be :

        * different (guid, weight) lists (possibly just order)
        * same length
        """
        assert (
            reduce(
                operator.and_,
                [
                    (t1[0] == t2[0] and t1[1] == t2[1])
                    for t1, t2 in zip(rand_list, raw_list)
                ],
            )
            is False
        )

        assert len(rand_list) == len(raw_list)



"""
These are global fixtures automagically loaded by pytest
"""

import pytest
from taar.context import _default_context
from taar.recommenders.redis_cache import TAARCacheRedis

FAKE_LOCALE_DATA = {
    "te-ST": [
        "{1e6b8bce-7dc8-481c-9f19-123e41332b72}",
        "some-other@nice-addon.com",
        "{66d1eed2-a390-47cd-8215-016e9fa9cc55}",
        "{5f1594c3-0d4c-49dd-9182-4fbbb25131a7}",
    ],
    "en": ["some-uuid@test-addon.com", "other-addon@some-id.it"],
}


@pytest.fixture(scope='function')
def test_ctx():
    ctx = _default_context(TAARCacheRedis)
    return ctx


@pytest.fixture
def TAARLITE_MOCK_DATA():
    return {
        "guid-1": {
            "guid-2": 1000,
            "guid-3": 100,
            "guid-4": 10,
            "guid-5": 1,
            "guid-6": 1,
        },
        "guid-2": {
            "guid-1": 50,
            "guid-3": 40,
            "guid-4": 20,
            "guid-8": 30,
            "guid-9": 10,
        },
        "guid-3": {"guid-1": 100, "guid-2": 40, "guid-4": 70},
        "guid-4": {"guid-2": 20},
        "guid-6": {"guid-1": 5, "guid-7": 100, "guid-8": 100, "guid-9": 100},
        "guid-8": {"guid-2": 30},
        "guid-9": {"guid-2": 10},
    }


@pytest.fixture
def TAARLITE_TIE_MOCK_DATA():
    return {
        "guid-1": {"guid-2": 100, "guid-3": 100, "guid-4": 100, "guid-5": 100},
        "guid-2": {"guid-1": 100, "guid-3": 100, "guid-4": 100, "guid-5": 100},
        "guid-3": {"guid-1": 100, "guid-2": 100, "guid-4": 100, "guid-5": 100},
        "guid-4": {"guid-1": 20, "guid-2": 20, "guid-3": 20, "guid-5": 20},
        "guid-5": {"guid-1": 20, "guid-2": 20, "guid-3": 20, "guid-4": 20},
    }


@pytest.fixture
def TAARLITE_MOCK_GUID_RANKING():
    return {
        "guid-1": 10,
        "guid-2": 9,
        "guid-3": 8,
        "guid-4": 7,
        "guid-5": 6,
        "guid-6": 5,
        "guid-7": 4,
        "guid-8": 3,
        "guid-9": 2,
    }


@pytest.fixture
def TAARLITE_CUTOFF_GUID_RANKING():
    return {
        "guid-1": 10000,
        "guid-2": 9000,
        "guid-3": 8000,
        "guid-4": 7,
        "guid-5": 6000,
        "guid-6": 5000,
        "guid-7": 4000,
        "guid-8": 3000,
        "guid-9": 2000,
    }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from flask import request
import json

import markus
from sentry_sdk import capture_exception

# TAAR specific libraries
from taar.context import default_context
from taar.logs.moz_logging import ContextFilter, Logging
from taar.profile_fetcher import ProfileFetcher
from taar.recommenders.guid_based_recommender import GuidBasedRecommender
from taar.recommenders.recommendation_manager import RecommenderFactory, RecommendationManager
from taar.recommenders.redis_cache import TAARCacheRedis
from taar.recommenders.cache import TAARCache

from taar.settings import (
    TAAR_MAX_RESULTS,
    TAARLITE_MAX_RESULTS,
    STATSD_HOST,
    STATSD_PORT,
    PYTHON_LOG_LEVEL,
    NO_REDIS
)


def acquire_taarlite_singleton(PROXY_MANAGER):
    if PROXY_MANAGER.getTaarLite() is None:
        cache_cls = TAARCache if NO_REDIS else TAARCacheRedis
        ctx = default_context(cache_cls=cache_cls, logger_cls=Logging, log_level=PYTHON_LOG_LEVEL)
        root_ctx = ctx.child()
        instance = GuidBasedRecommender(root_ctx)
        PROXY_MANAGER.setTaarLite(instance)
    return PROXY_MANAGER.getTaarLite()


def acquire_taar_singleton(PROXY_MANAGER):
    if PROXY_MANAGER.getTaarRM() is None:
        cache_cls = TAARCache if NO_REDIS else TAARCacheRedis
        ctx = default_context(cache_cls=cache_cls, logger_cls=Logging, log_level=PYTHON_LOG_LEVEL)

        profile_fetcher = ProfileFetcher(ctx)
        ctx["profile_fetcher"] = profile_fetcher

        # Lock the context down after we've got basic bits installed
        root_ctx = ctx.child()
        r_factory = RecommenderFactory(root_ctx)
        root_ctx["recommender_factory"] = r_factory
        instance = RecommendationManager(root_ctx.child())
        PROXY_MANAGER.setTaarRM(instance)
    return PROXY_MANAGER.getTaarRM()


class ResourceProxy(object):
    def __init__(self):
        self._resource = None
        self._taarlite_resource = None

    def setTaarRM(self, rsrc):
        self._resource = rsrc

    def getTaarRM(self):
        return self._resource

    def setTaarLite(self, rsrc):
        self._taarlite_resource = rsrc

    def getTaarLite(self):
        return self._taarlite_resource


PROXY_MANAGER = ResourceProxy()


def clean_promoted_guids(raw_promoted_guids):
    """ Verify that the promoted GUIDs are formatted correctly,
    otherwise strip it down into an empty list.
    """
    valid = True

    for row in raw_promoted_guids:
        if len(row) != 2:
            valid = False
            break

        if not (
            (isinstance(row[0], str))
            and (isinstance(row[1], int) or isinstance(row[1], float))  # noqa
        ):
            valid = False
            break

    if valid:
        return raw_promoted_guids
    return []


def merge_promoted_guids(promoted_guids, recommended_guids):
    guids = set()
    final = []
    tmp = sorted(
        promoted_guids + [x for x in recommended_guids],
        key=lambda x: x[1],
        reverse=True,
    )
    for guid, weight in tmp:
        if guid not in guids:
            final.append((guid, weight))
            guids.add(guid)
    return final


def configure_plugin(app):  # noqa: C901
    """
    This is a factory function that configures all the routes for
    flask given a particular library.
    """

    markus.configure(
        backends=[
            {
                # Log metrics to local instance of statsd
                # server. Use DatadogMetrics client
                "class": "markus.backends.datadog.DatadogMetrics",
                "options": {
                    "statsd_host": STATSD_HOST,
                    "statsd_port": STATSD_PORT,
                    "statsd_namespace": "",
                },
            }
        ]
    )

    @app.route("/taarlite/api/v1/addon_recommendations/<string:guid>/")
    def taarlite_recommendations(guid):
        """Return a list of recommendations provided a telemetry client_id."""
        # Use the module global PROXY_MANAGER
        global PROXY_MANAGER
        taarlite_recommender = acquire_taarlite_singleton(PROXY_MANAGER)

        cdict = {"guid": guid}
        normalization_type = request.args.get("normalize", None)
        if normalization_type is not None:
            cdict["normalize"] = normalization_type

        def set_extra(record):
            record.url = request.path
            record.guid = guid

        with ContextFilter(taarlite_recommender.logger, set_extra):
            recommendations = taarlite_recommender.recommend(
                client_data=cdict, limit=TAARLITE_MAX_RESULTS
            )

        if len(recommendations) != TAARLITE_MAX_RESULTS:
            recommendations = []

        # Strip out weights from TAAR results to maintain compatibility
        # with TAAR 1.0
        jdata = {"results": [x[0] for x in recommendations]}

        response = app.response_class(
            response=json.dumps(jdata), status=200, mimetype="application/json"
        )
        return response

    @app.route(
        "/v1/api/client_has_addon/<hashed_client_id>/<addon_id>/", methods=["GET"],
    )
    def client_has_addon(hashed_client_id, addon_id):
        # Use the module global PROXY_MANAGER
        global PROXY_MANAGER
        recommendation_manager = acquire_taar_singleton(PROXY_MANAGER)
        pf = recommendation_manager._ctx["profile_fetcher"]

        client_meta = pf.get(hashed_client_id)
        if client_meta is None:
            # no valid client metadata was found for the given
            # clientId
            result = {"results": False, "error": "No client found"}
            response = app.response_class(
                response=json.dumps(result), status=200, mimetype="application/json",
            )
            return response

        result = {"results": addon_id in client_meta.get("installed_addons", [])}
        response = app.response_class(
            response=json.dumps(result), status=200, mimetype="application/json"
        )
        return response

    @app.route("/v1/api/recommendations/<hashed_client_id>/", methods=["GET", "POST"])
    def recommendations(hashed_client_id):
        """Return a list of recommendations provided a telemetry client_id."""
        # Use the module global PROXY_MANAGER
        global PROXY_MANAGER

        extra_data = {}
        extra_data["options"] = {}
        extra_data["options"]["promoted"] = []

        try:
            if request.method == "POST":
                json_data = request.data
                # At least Python3.5 returns request.data as bytes
                # type instead of a string type.
                # Both Python2.7 and Python3.7 return a string type
                if type(json_data) == bytes:
                    json_data = json_data.decode("utf8")

                if json_data != "":
                    post_data = json.loads(json_data)
                    raw_promoted_guids = post_data.get("options", {}).get(
                        "promoted", []
                    )
                    promoted_guids = clean_promoted_guids(raw_promoted_guids)
                    extra_data["options"]["promoted"] = promoted_guids

        except Exception as e:
            jdata = {}
            jdata["results"] = []
            jdata["error"] = "Invalid JSON in POST: {}".format(e)
            capture_exception(e)
            return app.response_class(
                response=json.dumps(jdata, status=400, mimetype="application/json")
            )

        # Coerce the uuid.UUID type into a string
        client_id = str(hashed_client_id)

        locale = request.args.get("locale", None)
        if locale is not None:
            extra_data["locale"] = locale

        platform = request.args.get("platform", None)
        if platform is not None:
            extra_data["platform"] = platform

        recommendation_manager = acquire_taar_singleton(PROXY_MANAGER)

        def set_extra(record):
            record.url = request.path
            if locale:
                record.locale = locale
            if platform:
                record.platform = platform
            record.client_id = client_id
            record.method = request.method

        with ContextFilter(recommendation_manager.logger, set_extra):
            recommendations = recommendation_manager.recommend(
                client_id=client_id, limit=TAAR_MAX_RESULTS, extra_data=extra_data
            )

        promoted_guids = extra_data.get("options", {}).get("promoted", [])
        recommendations = merge_promoted_guids(promoted_guids, recommendations)

        # Strip out weights from TAAR results to maintain compatibility
        # with TAAR 1.0
        jdata = {"results": [x[0] for x in recommendations]}

        response = app.response_class(
            response=json.dumps(jdata), status=200, mimetype="application/json"
        )
        return response

    class MyPlugin:
        def set(self, config_options):
            """
            This setter is primarily so that we can instrument the
            cached RecommendationManager implementation under test.

            All plugins should implement this set method to enable
            overwriting configuration options with a TAAR library.
            """
            global PROXY_MANAGER
            if "PROXY_RESOURCE" in config_options:
                PROXY_MANAGER._resource = config_options["PROXY_RESOURCE"]

    return MyPlugin()



"""

Noop helpers
"""

import mock
from taar.recommenders.redis_cache import TAARCacheRedis


def noop_taarlite_dataload(stack):
    # no-op the taarlite rankdata
    stack.enter_context(
        mock.patch.object(TAARCacheRedis, "_update_rank_data", return_value=None)
    )
    # no-op the taarlite guidguid data
    stack.enter_context(
        mock.patch.object(TAARCacheRedis, "_update_coinstall_data", return_value=None, )
    )
    return stack


def noop_taarlocale_dataload(stack):
    # no-op the taarlite rankdata
    stack.enter_context(
        mock.patch.object(TAARCacheRedis, "_update_locale_data", return_value=None)
    )
    return stack


def noop_taarcollab_dataload(stack):
    # no-op the taar collab
    stack.enter_context(
        mock.patch.object(TAARCacheRedis, "_update_collab_data", return_value=None)
    )
    return stack


def noop_taarsimilarity_dataload(stack):
    # no-op the taar collab
    stack.enter_context(
        mock.patch.object(TAARCacheRedis, "_update_similarity_data", return_value=None)
    )
    return stack


def noop_taarensemble_dataload(stack):
    # no-op the taar collab
    stack.enter_context(
        mock.patch.object(TAARCacheRedis, "_update_ensemble_data", return_value=None)
    )
    stack.enter_context(
        mock.patch.object(TAARCacheRedis, "_update_whitelist_data", return_value=None)
    )
    return stack



from .collaborative_recommender import CollaborativeRecommender
from .locale_recommender import LocaleRecommender
from .similarity_recommender import SimilarityRecommender


__all__ = [
    "CollaborativeRecommender",
    "LocaleRecommender",
    "SimilarityRecommender",
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.logs.interfaces import IMozLogging, ITAARCache
import numpy as np
import operator as op

from taar.recommenders.base_recommender import AbstractRecommender


def java_string_hashcode(s):
    h = 0
    for c in s:
        h = (31 * h + ord(c)) & 0xFFFFFFFF
    return ((h + 0x80000000) & 0xFFFFFFFF) - 0x80000000


def positive_hash(s):
    return java_string_hashcode(s) & 0x7FFFFF


class CollaborativeRecommender(AbstractRecommender):
    """ The addon recommendation interface to the collaborative filtering model.

    Usage example::

        recommender = CollaborativeRecommender()
        dists = recommender.recommend(client_info)
    """

    def __init__(self, ctx):
        self._ctx = ctx

        self.logger = self._ctx[IMozLogging].get_logger("taar")

        self._cache = self._ctx[ITAARCache]

    def _get_cache(self, extra_data):
        tmp = extra_data.get("cache", None)
        if tmp is None:
            tmp = self._cache.cache_context()
        return tmp

    def can_recommend(self, client_data, extra_data={}):
        cache = self._get_cache(extra_data)
        # We can't recommend if we don't have our data files.
        if (
                cache["raw_item_matrix"] is None
                or cache["collab_model"] is None
                or cache["addon_mapping"] is None
        ):
            return False

        # We only get meaningful recommendation if a client has at least an
        # addon installed.
        if len(client_data.get("installed_addons", [])) > 0:
            return True

        return False

    def _recommend(self, client_data, limit, extra_data):
        cache = self._get_cache(extra_data)

        installed_addons_as_hashes = [
            positive_hash(addon_id)
            for addon_id in client_data.get("installed_addons", [])
        ]

        # Build the query vector by setting the position of the queried addons to 1.0
        # and the other to 0.0.
        query_vector = np.array(
            [
                1.0 if (entry.get("id") in installed_addons_as_hashes) else 0.0
                for entry in cache["raw_item_matrix"]
            ]
        )

        # Build the user factors matrix.
        user_factors = np.matmul(query_vector, cache["collab_model"])
        user_factors_transposed = np.transpose(user_factors)

        # Compute the distance between the user and all the addons in the latent
        # space.
        distances = {}
        for addon in cache["raw_item_matrix"]:
            # We don't really need to show the items we requested.
            # They will always end up with the greatest score. Also
            # filter out legacy addons from the suggestions.
            hashed_id = addon.get("id")
            str_hashed_id = str(hashed_id)
            if (
                    hashed_id in installed_addons_as_hashes
                    or str_hashed_id not in cache["addon_mapping"]
                    or cache["addon_mapping"][str_hashed_id].get("isWebextension", False)
                    is False
            ):
                continue

            dist = np.dot(user_factors_transposed, addon.get("features"))
            # Read the addon ids from the "addon_mapping" looking it
            # up by 'id' (which is an hashed value).
            addon_id = cache["addon_mapping"][str_hashed_id].get("id")
            distances[addon_id] = dist

        # Sort the suggested addons by their score and return the
        # sorted list of addon ids.
        sorted_dists = sorted(distances.items(), key=op.itemgetter(1), reverse=True)
        recommendations = [(s[0], s[1]) for s in sorted_dists[:limit]]
        return recommendations

    def recommend(self, client_data, limit, extra_data={}):
        # Addons identifiers are stored as positive hash values within the model.

        recommendations = self._recommend(client_data, limit, extra_data)

        log_data = (
            client_data["client_id"],
            str([r[0] for r in recommendations]),
        )
        self.logger.debug(
            "collaborative_recommender_triggered, "
            "client_id: [%s], "
            "guids: [%s]" % log_data
        )

        return recommendations



import json

import fakeredis
import pytest
import mock
import contextlib

from taar.logs.interfaces import ITAARCache
from .noop_fixtures import (
    noop_taarlocale_dataload,
    noop_taarcollab_dataload,
    noop_taarsimilarity_dataload,
    noop_taarensemble_dataload,
)

from taar.recommenders.guid_based_recommender import GuidBasedRecommender
from taar.recommenders.redis_cache import TAARCacheRedis

from taar.recommenders.cache import NORMDATA_GUID_ROW_NORM_PREFIX

from taar.recommenders.ua_parser import parse_ua, OSNAME_TO_ID

from markus import TIMING
from markus.testing import MetricsMock

# The different kinds of results we can expect from TAARlite are
# listed below.  Note that the ordering of GUIDs returned, and even
# the set of GUIDs returned may be altered by the different weight
# normalization modes.
#
# Based on some preliminary clustering analysis, the 'rownorm_sum'
# method appears to provide qualitatively better results than the
# other normalization modes including no normalization.


# Reading the RESULTS is not entirely obvious.  The recommendation
# list consists of 2-tuples containing a guid, followed by a lexically
# sorted weight+install count.
# The weights are formatted as a fixed with zero padded float, with
# an addition suffix of a decimal and a zero padded instllation count
# for the addon.
#
# The clearest example of this is the 'rownorm_sum_tiebreak' results
# where each of the computed weights are the same (0.25), but the
# installation count varies.
RESULTS = {
    "default": [
        ("guid-2", "000001000.0000000000.0000000009"),
        ("guid-3", "000000100.0000000000.0000000008"),
        ("guid-4", "000000010.0000000000.0000000007"),
        ("guid-5", "000000001.0000000000.0000000006"),
    ],
    "row_count": [
        ("guid-3", "000000020.0000000000.0000000008"),  # 50% of 40
        ("guid-1", "000000016.6666666667.0000000010"),  # 1/3 of 50
        ("guid-8", "000000015.0000000000.0000000003"),  # 50% of 30
        ("guid-4", "000000006.6666666667.0000000007"),
    ],  # 1/3 of 20
    "rownorm_sum": [
        ("guid-3", "000000000.7478143914.0000000008"),
        ("guid-1", "000000000.4035916824.0000000010"),
        ("guid-8", "000000000.3788819876.0000000003"),
        ("guid-4", "000000000.2803125788.0000000007"),
    ],
    "rownorm_sum_tiebreak": [
        ("guid-1", "000000000.2500000000.0000000010"),
        ("guid-3", "000000000.2500000000.0000000008"),
        ("guid-4", "000000000.2500000000.0000000007"),
        ("guid-5", "000000000.2500000000.0000000006"),
    ],
    "rownorm_sum_tiebreak_cutoff": [
        ("guid-1", "000000000.3333333333.0000010000"),  # noqa
        ("guid-3", "000000000.3333333333.0000008000"),  # noqa
        ("guid-5", "000000000.3333333333.0000006000"),
    ],  # noqa
    "row_sum": [
        ("guid-1", "000000000.3225806452.0000000010"),
        ("guid-3", "000000000.2857142857.0000000008"),
        ("guid-8", "000000000.2307692308.0000000003"),
        ("guid-4", "000000000.2000000000.0000000007"),
    ],
    "guidception": [
        ("guid-1", "000000000.2666666667.0000000010"),
        ("guid-3", "000000000.2333333333.0000000008"),
        ("guid-8", "000000000.2000000000.0000000003"),
        ("guid-4", "000000000.1666666667.0000000007"),
    ],
}


@contextlib.contextmanager
def mock_coinstall_ranking_context(ctx, mock_coinstall, mock_ranking):

    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None

        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_ranking_data", return_value=mock_ranking,
            )
        )
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_coinstall_data", return_value=mock_coinstall,
            )
        )

        stack = noop_taarlocale_dataload(stack)
        stack = noop_taarcollab_dataload(stack)
        stack = noop_taarsimilarity_dataload(stack)
        stack = noop_taarensemble_dataload(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


def test_recommender_nonormal(test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):
        with MetricsMock() as mm:
            EXPECTED_RESULTS = RESULTS["default"]
            recommender = GuidBasedRecommender(test_ctx)

            guid = "guid-1"

            actual = recommender.recommend({"guid": guid, "normalize": "none"})
            assert actual == EXPECTED_RESULTS

            mm.has_record(TIMING, "taar.guid_coinstall")
            mm.has_record(TIMING, "taar.guid_ranking")
            mm.has_record(TIMING, "taar.guid_recommendation")


def test_row_count_recommender(
    test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):

        EXPECTED_RESULTS = RESULTS["row_count"]

        recommender = GuidBasedRecommender(test_ctx)
        guid = "guid-2"

        actual = recommender.recommend({"guid": guid, "normalize": "row_count"})

        # Note that guid-9 is not included because it's weight is
        # decreased 50% to 5
        assert EXPECTED_RESULTS == actual


def test_rownorm_sumrownorm(test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):
        EXPECTED_RESULTS = RESULTS["rownorm_sum"]
        recommender = GuidBasedRecommender(test_ctx)
        guid = "guid-2"

        default_actual = recommender.recommend({"guid": guid})

        actual = recommender.recommend({"guid": guid, "normalize": "rownorm_sum"})

        # Default normalization is rownorm_sum
        assert actual == default_actual
        assert actual == EXPECTED_RESULTS
        """
        Some notes on verifying guid-1:

        Numerator is the row weighted value of guid-1 : 50/150
        Denominator is the sum of the row weighted value of guid-1 in all
        other rows

        (guid-2) 50/150
        (guid-3) 100/210
        (guid-6) 5/305

        This gives us: [0.3333333333333333,
                        0.47619047619047616,
                        0.01639344262295082]

        so the final result should be (5/150) / (50/150 + 100/210 + 5/305)

        That gives a final expected weight for guid-1 to be: 0.403591682
        """
        expected = 0.403591682
        actual = float(actual[1][1][:-11])
        assert expected == pytest.approx(actual, rel=1e-3)


def test_rowsum_recommender(test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):
        EXPECTED_RESULTS = RESULTS["row_sum"]

        recommender = GuidBasedRecommender(test_ctx)
        guid = "guid-2"

        actual = recommender.recommend({"guid": guid, "normalize": "row_sum"})
        assert 4 == len(actual)

        expected_val = 50 / 155
        actual_val = float(actual[0][1][:-11])
        assert expected_val == pytest.approx(actual_val, rel=1e-3)

        assert actual == EXPECTED_RESULTS


def test_guidception(test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):

        EXPECTED_RESULTS = RESULTS["guidception"]

        recommender = GuidBasedRecommender(test_ctx)
        guid = "guid-2"

        actual = recommender.recommend({"guid": guid, "normalize": "guidception"})
        assert actual == EXPECTED_RESULTS


def test_rownorm_sum_tiebreak(
    test_ctx, TAARLITE_TIE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_TIE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):
        EXPECTED_RESULTS = RESULTS["rownorm_sum_tiebreak"]

        recommender = GuidBasedRecommender(test_ctx)
        guid = "guid-2"

        actual = recommender.recommend({"guid": guid, "normalize": "rownorm_sum"})

        # Note that the results have weights that are equal, but the tie
        # break is solved by the install rate.
        assert actual == EXPECTED_RESULTS


def test_missing_rownorm_data_issue_31(
    test_ctx, TAARLITE_TIE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_TIE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):

        recommender = GuidBasedRecommender(test_ctx)

        EXPECTED_RESULTS = RESULTS["rownorm_sum_tiebreak"]

        # Explicitly destroy the guid-4 key in the row_norm data
        recommender._cache._db().delete(NORMDATA_GUID_ROW_NORM_PREFIX + "guid-4")
        for i, row in enumerate(EXPECTED_RESULTS):
            if row[0] == "guid-4":
                del EXPECTED_RESULTS[i]
                break

        guid = "guid-2"

        actual = recommender.recommend({"guid": guid, "normalize": "rownorm_sum"})

        assert actual == EXPECTED_RESULTS


def test_divide_by_zero_rownorm_data_issue_31(
    test_ctx, TAARLITE_TIE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
):
    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_TIE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):

        recommender = GuidBasedRecommender(test_ctx)

        EXPECTED_RESULTS = RESULTS["rownorm_sum_tiebreak"]

        # Explicitly set the guid-4 key in the row_norm data to have a sum
        # of zero weights
        recommender._cache._db().set(
            NORMDATA_GUID_ROW_NORM_PREFIX + "guid-4", json.dumps([0, 0, 0])
        )

        # Destroy the guid-4 key in the expected results as a sum of 0
        # will generate a divide by zero error
        for i, row in enumerate(EXPECTED_RESULTS):
            if row[0] == "guid-4":
                del EXPECTED_RESULTS[i]
                break

        guid = "guid-2"

        actual = recommender.recommend({"guid": guid, "normalize": "rownorm_sum"})

        assert actual == EXPECTED_RESULTS


def test_user_agent_strings():
    """
    The UA String parser should only care about selecting the right
    platform for Firefox UA strings.  Any non-firefox browser should
    get all available addons.
    """
    ua_strings = {
        "windows": "Mozilla/5.0 (Windows NT x.y; rv:10.0) Gecko/20100101 Firefox/10.0",
        "macintosh": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:60.0) Gecko/20100101 Firefox/60.0",
        "linux": "Mozilla/5.0 (X11; Linux i686; rv:10.0) Gecko/20100101 Firefox/10.0",
        "android": "Mozilla/5.0 (Android; Mobile; rv:40.0) Gecko/40.0 Firefox/40.0",
    }

    not_fx_ua_strings = [
        # Chrome
        "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.36",  # noqa
        # Microsoft Edge
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.246",  # noqa
        # Safari
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.75.14 (KHTML, like Gecko) Version/7.0.3 Safari/7046A194A",  # noqa
    ]

    # These are valid Firefox UA strings
    for platform, ua in ua_strings.items():
        platform_id = parse_ua(ua)
        assert OSNAME_TO_ID[platform] == platform_id

    # These are non-Firefox UA strings - we should expect nothing
    for ua in not_fx_ua_strings:
        actual_name = parse_ua(ua)
        assert actual_name is None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# copy paste from https://github.com/mozilla/srgutil to get rid of this heavy legacy dependency
try:
    from abc import ABC
except Exception:
    from abc import ABCMeta

    class ABC(object):
        """Helper class that provides a standard way to create an ABC using
        inheritance.
        """
        __metaclass__ = ABCMeta
        __slots__ = ()


class IMozLogging(ABC):
    def get_logger(self, name):
        """Get a logger with the current configuration
        """

    def set_log_level(self, level):
        """Set the logs level, fox example 'DEBUG'
        """


class ITAARCache(ABC):
    def safe_load_data(self):
        raise NotImplementedError()

    def cache_context(self):
        raise NotImplementedError()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from taar.logs.interfaces import IMozLogging, ITAARCache

from .base_recommender import AbstractRecommender


class LocaleRecommender(AbstractRecommender):
    """ A recommender class that returns top N addons based on the client geo-locale.

    This will load a json file containing updated top n addons in use per geo locale
    updated periodically by a separate process on airflow using Longitdudinal Telemetry
    data.

    This recommender may provide useful recommendations when collaborative_recommender
    may not work.
    """

    def __init__(self, ctx):
        self._ctx = ctx

        self.logger = self._ctx[IMozLogging].get_logger("taar")

        self._cache = self._ctx[ITAARCache]

    def _get_cache(self, extra_data):
        tmp = extra_data.get("cache", None)
        if tmp is None:
            tmp = self._cache.cache_context()
        return tmp

    def can_recommend(self, client_data, extra_data={}):
        cache = self._get_cache(extra_data)

        # We can't recommend if we don't have our data files.
        if cache["top_addons_per_locale"] is None:
            return False

        # If we have data coming from other sources, we can use that for
        # recommending.
        client_locale = client_data.get("locale", None) or extra_data.get(
            "locale", None
        )
        if not isinstance(client_locale, str):
            return False

        if client_locale not in cache["top_addons_per_locale"]:
            return False

        if not cache["top_addons_per_locale"].get(client_locale):
            return False

        return True

    def recommend(self, client_data, limit, extra_data={}):
        cache = self._get_cache(extra_data)
        # If we have data coming from multiple sourecs, prefer the one
        # from 'client_data'.
        client_locale = client_data.get("locale") or extra_data.get("locale", None)
        result_list = cache["top_addons_per_locale"].get(client_locale, [])[:limit]

        if "locale" not in client_data:
            try:
                client_data["locale"] = extra_data["locale"]
            except KeyError:
                client_data["locale"] = None

        log_data = (client_data["locale"], str([r[0] for r in result_list]))
        self.logger.debug(
            "locale_recommender_triggered, "
            "client_locale: [%s], guids: [%s]" % log_data
        )
        return result_list



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
from flask import Flask
from dockerflow.flask import Dockerflow
import optparse
from decouple import config
import importlib
import sentry_sdk
from sentry_sdk.integrations.flask import FlaskIntegration


app = Flask(__name__)
dockerflow = Dockerflow(app)

# Hook the application plugin and configure it
PLUGIN = config("TAAR_API_PLUGIN", default=None)


sentry_sdk.init(
    dsn=config("SENTRY_DSN", ""), integrations=[FlaskIntegration()],
)

# There should only be a single registered app for the taar-api
if PLUGIN is None:
    sys.stderr.write("No plugin is defined.\n")
    sys.exit(1)


# Load the function and configure the application
sys.stdout.write("Loading [{}]\n".format(PLUGIN))

plugin_module = importlib.import_module(PLUGIN)
configure_plugin = importlib.import_module(PLUGIN).configure_plugin
APP_WRAPPER = configure_plugin(app)


def flaskrun(app, default_host="127.0.0.1", default_port="8000"):
    """
    Takes a flask.Flask instance and runs it. Parses
    command-line flags to configure the app.
    """

    # Set up the command-line options
    parser = optparse.OptionParser()
    parser.add_option(
        "-H",
        "--host",
        help="Hostname of the Flask app " + "[default %s]" % default_host,
        default=default_host,
    )
    parser.add_option(
        "-P",
        "--port",
        help="Port for the Flask app " + "[default %s]" % default_port,
        default=default_port,
    )

    # Two options useful for debugging purposes, but
    # a bit dangerous so not exposed in the help message.
    parser.add_option(
        "-d", "--debug", action="store_true", dest="debug", help=optparse.SUPPRESS_HELP
    )
    parser.add_option(
        "-p",
        "--profile",
        action="store_true",
        dest="profile",
        help=optparse.SUPPRESS_HELP,
    )

    options, _ = parser.parse_args()

    # If the user selects the profiling option, then we need
    # to do a little extra setup
    if options.profile:
        from werkzeug.contrib.profiler import ProfilerMiddleware

        app.config["PROFILE"] = True
        app.wsgi_app = ProfilerMiddleware(app.wsgi_app, restrictions=[30])
        options.debug = True

    app.run(debug=options.debug, host=options.host, port=int(options.port))


if __name__ == "__main__":
    flaskrun(app)



# Each of these records is identical with respect to the
# CATEGORICAL_FEATURES and CONTINUOUS_FEATURES of the TAAR client
# generated by `generate_a_fake_taar_client` except for the
# `total_uri` field.  This gives us a deterministic way to order the
# recommendations provided by the SimilarityRecommender based on
# continuous feature similarity.

CONTINUOUS_FEATURE_FIXTURE_DATA = [
    {
        "active_addons": [
            "{test-guid-1}",
            "{test-guid-2}",
            "{test-guid-3}",
            "{test-guid-4}",
        ],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 190,
        "unique_tlds": 21,
    },
    {
        "active_addons": [
            "{test-guid-5}",
            "{test-guid-6}",
            "{test-guid-1}",
            "{test-guid-8}",
        ],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 200,
        "unique_tlds": 21,
    },
    {
        "active_addons": [
            "{test-guid-9}",
            "{test-guid-10}",
            "{test-guid-11}",
            "{test-guid-12}",
        ],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 222,
        "unique_tlds": 21,
    },
    {
        "active_addons": ["{test-guid-13}", "{test-guid-14}"],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 210,
        "unique_tlds": 21,
    },
]

# Match the fixture taar client, but vary the geo_city to test only
# the categorical feature matching.

# Additionally the second donor contains the only duplicate recommendation
# of "{test-guid-1}"

CATEGORICAL_FEATURE_FIXTURE_DATA = [
    {
        "active_addons": [
            "{test-guid-1}",
            "{test-guid-2}",
            "{test-guid-3}",
            "{test-guid-4}",
        ],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 222,
        "unique_tlds": 21,
    },
    {
        # "{test-guid-1}" appears in duplicate here.
        "active_addons": [
            "{test-guid-5}",
            "{test-guid-6}",
            "{test-guid-1}",
            "{test-guid-8}",
        ],
        "geo_city": "toronto-ca",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 222,
        "unique_tlds": 21,
    },
    {
        "active_addons": [
            "{test-guid-9}",
            "{test-guid-10}",
            "{test-guid-11}",
            "{test-guid-12}",
        ],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 222,
        "unique_tlds": 21,
    },
    {
        "active_addons": ["{test-guid-13}", "{test-guid-1}"],
        "geo_city": "toronto-ca",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 222,
        "unique_tlds": 21,
    },
]



import hashlib
import logging
import uuid

from flask import Flask
from flask import url_for

import pytest

from taar.recommenders.redis_cache import TAARCacheRedis
from taar.settings import TAARLITE_MAX_RESULTS
from taar.context import default_context
from .test_guid_based_recommender import mock_coinstall_ranking_context

try:
    from unittest.mock import MagicMock
except Exception:
    from mock import MagicMock


def hasher(uuid):
    return hashlib.new("sha256", str(uuid).encode("utf8")).hexdigest()


@pytest.fixture
def app():

    from taar.plugin import configure_plugin
    from taar.plugin import PROXY_MANAGER

    flask_app = Flask("test")

    # Clobber the default recommendation manager with a MagicMock
    mock_recommender = MagicMock()
    PROXY_MANAGER.setTaarRM(mock_recommender)

    configure_plugin(flask_app)

    return flask_app


def test_empty_results_by_default(client, app):
    # The default behaviour under test should be that the
    # RecommendationManager simply no-ops everything so we get back an
    # empty result list.
    res = client.post("/v1/api/recommendations/not_a_real_hash/")
    assert res.json == {"results": []}


def test_only_promoted_addons_post(client, app):
    # POSTing a JSON blob allows us to specify promoted addons to the
    # TAAR service.
    res = client.post(
        "/v1/api/recommendations/not_a_real_hash/",
        json=dict(
            {"options": {"promoted": [["guid1", 10], ["guid2", 5], ["guid55", 8]]}}
        ),
        follow_redirects=True,
    )
    # The result should order the GUIDs in descending order of weight
    assert res.json == {"results": ["guid1", "guid55", "guid2"]}


class FakeRecommendationManager(object):
    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger('test')


class StaticRecommendationManager(FakeRecommendationManager):

    # Recommenders must return a list of 2-tuple results
    # with (GUID, weight)
    def recommend(self, client_id, limit, extra_data={}):
        result = [
            ("test-addon-1", 1.0),
            ("test-addon-2", 1.0),
            ("test-addon-N", 1.0),
        ]
        return result


class LocaleRecommendationManager(FakeRecommendationManager):
    def recommend(self, client_id, limit, extra_data={}):
        if extra_data.get("locale", None) == "en-US":
            return [("addon-Locale", 1.0)]
        return []


class EmptyRecommendationManager(FakeRecommendationManager):
    def recommend(self, client_id, limit, extra_data={}):
        return []


class PlatformRecommendationManager(FakeRecommendationManager):
    def recommend(self, client_id, limit, extra_data={}):
        if extra_data.get("platform", None) == "WOW64":
            return [("addon-WOW64", 1.0)]
        return []


class ProfileFetcherEnabledRecommendationManager(FakeRecommendationManager):
    def __init__(self, *args, **kwargs):
        self._ctx = default_context(TAARCacheRedis)
        self._ctx["profile_fetcher"] = kwargs["profile_fetcher"]
        super(ProfileFetcherEnabledRecommendationManager, self).__init__(args, kwargs)


@pytest.fixture
def locale_recommendation_manager(monkeypatch):
    # Force the plugin configuration
    import os

    os.environ["TAAR_API_PLUGIN"] = "taar.plugin"

    import taar.flask_app

    taar.flask_app.APP_WRAPPER.set({"PROXY_RESOURCE": LocaleRecommendationManager()})


@pytest.fixture
def empty_recommendation_manager(monkeypatch):
    # Force the plugin configuration
    import os

    os.environ["TAAR_API_PLUGIN"] = "taar.plugin"

    import taar.flask_app

    taar.flask_app.APP_WRAPPER.set({"PROXY_RESOURCE": EmptyRecommendationManager()})


@pytest.fixture
def platform_recommendation_manager(monkeypatch):
    # Force the plugin configuration
    import os

    os.environ["TAAR_API_PLUGIN"] = "taar.plugin"

    import taar.flask_app

    taar.flask_app.APP_WRAPPER.set({"PROXY_RESOURCE": PlatformRecommendationManager()})


@pytest.fixture
def static_recommendation_manager(monkeypatch):
    # Force the plugin configuration
    import os

    os.environ["TAAR_API_PLUGIN"] = "taar.plugin"

    import taar.flask_app

    taar.flask_app.APP_WRAPPER.set({"PROXY_RESOURCE": StaticRecommendationManager()})


@pytest.fixture
def profile_enabled_rm(monkeypatch):
    # Force the plugin configuration
    import os

    os.environ["TAAR_API_PLUGIN"] = "taar.plugin"

    import taar.flask_app

    mock_profile = {"installed_addons": ["addon_119", "addon_219"]}

    class MockPF:
        def get(self, hashed_client_id):
            return mock_profile

    pf = MockPF()
    pfm = ProfileFetcherEnabledRecommendationManager(profile_fetcher=pf)
    taar.flask_app.APP_WRAPPER.set({"PROXY_RESOURCE": pfm})


def test_empty_recommendation(client, empty_recommendation_manager):
    response = client.post(
        url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
    )
    assert response.status_code == 200
    assert response.headers["Content-Type"] == "application/json"
    assert response.data == b'{"results": []}'


def test_locale_recommendation(client, locale_recommendation_manager):
    response = client.post(
        url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
        + "?locale=en-US"
    )
    assert response.status_code == 200
    assert response.headers["Content-Type"] == "application/json"
    assert response.data == b'{"results": ["addon-Locale"]}'

    response = client.post(
        url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
    )
    assert response.status_code == 200
    assert response.headers["Content-Type"] == "application/json"
    assert response.data == b'{"results": []}'


def test_platform_recommendation(client, platform_recommendation_manager):
    uri = (
        url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
        + "?platform=WOW64"
    )
    response = client.post(uri)
    assert response.status_code == 200
    assert response.headers["Content-Type"] == "application/json"
    assert response.data == b'{"results": ["addon-WOW64"]}'

    response = client.post(
        url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
    )
    assert response.status_code == 200
    assert response.headers["Content-Type"] == "application/json"
    assert response.data == b'{"results": []}'


def test_simple_request(client, static_recommendation_manager):
    url = url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
    response = client.post(url)
    assert response.status_code == 200
    assert response.headers["Content-Type"] == "application/json"
    expected = b'{"results": ["test-addon-1", "test-addon-2", "test-addon-N"]}'
    assert response.data == expected


def test_mixed_and_promoted_and_taar_adodns(client, static_recommendation_manager):
    """
    Test that we can provide addon suggestions that also get clobbered
    by the promoted addon set.
    """
    url = url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
    res = client.post(
        url,
        json=dict(
            {"options": {"promoted": [["guid1", 10], ["guid2", 5], ["guid55", 8]]}}
        ),
        follow_redirects=True,
    )
    # The result should order the GUIDs in descending order of weight
    expected = {
        "results": [
            "guid1",
            "guid55",
            "guid2",
            "test-addon-1",
            "test-addon-2",
            "test-addon-N",
        ]
    }
    assert res.json == expected


def test_overlapping_mixed_and_promoted_and_taar_adodns(
    client, static_recommendation_manager
):
    """
    Test that we can provide addon suggestions that also get clobbered
    by the promoted addon set.
    """
    url = url_for("recommendations", hashed_client_id=hasher(uuid.uuid4()))
    res = client.post(
        url,
        json=dict(
            {
                "options": {
                    "promoted": [["test-addon-1", 10], ["guid2", 5], ["guid55", 8],]
                }
            }
        ),
        follow_redirects=True,
    )
    # The result should order the GUIDs in descending order of weight
    expected = {
        "results": ["test-addon-1", "guid55", "guid2", "test-addon-2", "test-addon-N",]
    }
    assert res.json == expected


def test_client_addon_lookup_no_client(client, profile_enabled_rm):
    """
    test that we can see if a client has an addon installed
    """
    hashed_client_id = hasher(uuid.uuid4())
    addon_id = "abc123"

    url = url_for(
        "client_has_addon", hashed_client_id=hashed_client_id, addon_id=addon_id
    )
    res = client.get(url, follow_redirects=True)

    _ = {"results": False}
    assert res.json["results"] is False


def test_client_has_addon(client, profile_enabled_rm):
    """
    test that we can see if a client has an addon installed
    """

    hashed_client_id = hasher(uuid.uuid4())
    addon_id = "addon_119"

    url = url_for(
        "client_has_addon", hashed_client_id=hashed_client_id, addon_id=addon_id
    )
    res = client.get(url, follow_redirects=True)

    expected = {"results": True}
    assert res.json == expected


def test_client_has_no_addon(client, profile_enabled_rm):
    """
    test that we can see if a client has an addon installed
    """

    hashed_client_id = hasher(uuid.uuid4())
    addon_id = "addon_984932434"

    url = url_for(
        "client_has_addon", hashed_client_id=hashed_client_id, addon_id=addon_id
    )
    res = client.get(url, follow_redirects=True)

    assert res.json["results"] is False


def test_taarlite(client, test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING):
    """
    Check that the result size of taarlite is TAARLITE_MAX_RESULTS
    """

    with mock_coinstall_ranking_context(
        test_ctx, TAARLITE_MOCK_DATA, TAARLITE_MOCK_GUID_RANKING
    ):

        url = url_for("taarlite_recommendations", guid="guid-1",)
        res = client.get(url, follow_redirects=True)

        assert len(res.json["results"]) == TAARLITE_MAX_RESULTS
        assert res.json["results"] == ["guid-5", "guid-6", "guid-3", "guid-2"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.


from taar.logs.interfaces import IMozLogging, ITAARCache

import markus

from taar.recommenders.debug import log_timer_debug

metrics = markus.get_metrics("taar")


NORM_MODE_ROWNORMSUM = "rownorm_sum"
NORM_MODE_ROWCOUNT = "row_count"
NORM_MODE_ROWSUM = "row_sum"
NORM_MODE_GUIDCEPTION = "guidception"


class GuidBasedRecommender:
    """ A recommender class that returns top N addons based on a
    passed addon identifier.  This will load a json file containing
    updated top n addons coinstalled with the addon passed as an input
    parameter based on periodically updated  addon-addon
    coinstallation frequency table generated from  Longitdudinal
    Telemetry data.  This recommender will drive recommendations
    surfaced on addons.mozilla.org


    We store the JSON data for the GUID coinstallation in memory. This
    consumes ~ 15.8MB of heap.

        In [10]: from pympler import asizeof

        In [11]: jdata = json.load(open('guid_coinstallation.json'))

        In [12]: asizeof.asizeof(jdata)
        Out[12]: 15784672

    Each of the data normalization dictionaries is also stored in
    memory.
    """

    _addons_coinstallations = None
    _guid_maps = {}

    # Define recursion levels for guid-ception
    RECURSION_LEVELS = 3

    def __init__(self, ctx):
        self._ctx = ctx
        self.logger = self._ctx[IMozLogging].get_logger("taarlite")

        self._cache = ctx[ITAARCache]
        self.logger.info("GUIDBasedRecommender is initialized")

    def cache_ready(self):
        return self._cache.is_active()

    def can_recommend(self, client_data):
        # We can't recommend if we don't have our data files.
        if not self._cache.is_active():
            return False

        # If we have data coming from other sources, we can use that for
        # recommending.
        addon_guid = client_data.get("guid", None)
        if not isinstance(addon_guid, str):
            return False

        # Use a dictionary keyed on the query guid
        if not self._cache.has_coinstalls_for(addon_guid):
            return False

        if not self._cache.get_coinstalls(addon_guid):
            return False

        return True

    @metrics.timer_decorator("guid_recommendation")
    def recommend(self, client_data, limit=4):
        """
        TAAR lite will yield 4 recommendations for the AMO page
        """

        if not self._cache.is_active():
            return []

        with log_timer_debug(f"Results computed", self.logger):

            with log_timer_debug("get client data", self.logger):
                addon_guid = client_data.get("guid")

            # Get the raw co-installation result dictionary
            with log_timer_debug("Get filtered coinstallations", self.logger):
                result_dict = self._cache.get_filtered_coinstall(addon_guid, {})

            with log_timer_debug("acquire normalization method", self.logger):
                normalize = client_data.get("normalize", NORM_MODE_ROWNORMSUM)

                norm_dict = {
                    "none": lambda guid, x: x,
                    NORM_MODE_ROWCOUNT: self.norm_row_count,
                    NORM_MODE_ROWSUM: self.norm_row_sum,
                    NORM_MODE_ROWNORMSUM: self.norm_rownorm_sum,
                    NORM_MODE_GUIDCEPTION: self.norm_guidception,
                }

                if normalize is not None and normalize not in norm_dict.keys():
                    # Yield no results if the normalization method is not
                    # specified
                    self.logger.warning(
                        "Invalid normalization parameter detected: [%s]" % normalize
                    )
                    return []

                # Bind the normalization method
                norm_method = norm_dict[normalize]

            with log_timer_debug(
                f"Compute normalization using method:{normalize}", self.logger
            ):
                # Apply normalization
                tmp_result_dict = norm_method(addon_guid, result_dict)

            # Augment the result_dict with the installation counts
            # and then we can sort using lexical sorting of strings.
            # The idea here is to get something in the form of
            #    0000.0000.0000
            # The computed weight takes the first and second segments of
            # integers.  The third segment is the installation count of
            # the addon but is zero padded.

            TWICE_LIMIT = limit * 2
            with log_timer_debug(
                f"Augment {TWICE_LIMIT} with installation counts and resorted",
                self.logger,
            ):
                result_list = []
                rank_sorted = sorted(
                    tmp_result_dict.items(), key=lambda x: x[1], reverse=True
                )
                for k, v in rank_sorted[:TWICE_LIMIT]:
                    lex_value = "{0:020.10f}.{1:010d}".format(
                        v, self._cache.get_rankings(k, 0)
                    )
                    result_list.append((k, lex_value))

                # Sort the result list in descending order by weight
                result_list.sort(key=lambda x: x[1], reverse=True)

            self.logger.info(
                "Addon related recommendations results",
                extra={'guid': str(addon_guid), 'recs': [str(r[0]) for r in result_list[:limit]]}
            )

        return result_list[:limit]

    def norm_row_count(self, key_guid, input_coinstall_dict):
        """This normalization method counts the unique times that a
        GUID is coinstalled with any other GUID.

        This dampens weight of any suggested GUID inversely
        proportional to it's overall popularity.
        """

        output_result_dict = {}
        for result_guid, result_count in input_coinstall_dict.items():
            output_result_dict[result_guid] = (
                1.0 * result_count / self._cache.guid_maps_rowcount(result_guid)
            )
        return output_result_dict

    def norm_row_sum(self, key_guid, input_coinstall_dict):
        """This normalization normalizes the weights for the suggested
        coinstallation GUIDs based on the sum of the weights for the
        coinstallation GUIDs given a key GUID.
        """

        def generate_row_sum_list():
            for guid, guid_weight in input_coinstall_dict.items():
                norm_guid_weight = (
                    guid_weight * 1.0 / self._cache.guid_maps_count_map(guid)
                )

                yield guid, norm_guid_weight

        return dict(generate_row_sum_list())

    def norm_rownorm_sum(self, key_guid, input_coinstall_dict):
        """This normalization is the same as norm_row_sum, but we also
        divide the result by the sum of
        (addon coinstall instances)/(addon coinstall total instances)

        The testcase for this scenario lays out the math more
        explicitly.
        """
        with log_timer_debug("normalize row weights for coinstall dict", self.logger):
            tmp_dict = self._normalize_row_weights(input_coinstall_dict)

        with log_timer_debug(
            f"normalizing output_dict of size: {len(tmp_dict)}", self.logger
        ):
            output_dict = {}
            for output_guid, output_guid_weight in tmp_dict.items():
                guid_row_norm_list = self._cache.guid_maps_row_norm(
                    output_guid, []
                )
                if len(guid_row_norm_list) == 0:
                    self.logger.warning(
                        "Can't find GUID_ROW_NORM data for [{}]".format(output_guid)
                    )
                    continue
                norm_sum = sum(guid_row_norm_list)
                if norm_sum == 0:
                    self.logger.warning(
                        "Sum of GUID_ROW_NORM data for [{}] is zero.".format(
                            output_guid
                        )
                    )
                    continue
                output_dict[output_guid] = output_guid_weight / norm_sum

        return output_dict

    def norm_guidception(self, key_guid, input_coinstall_dict):
        tmp_dict = self._normalize_row_weights(input_coinstall_dict)

        return self._compute_recursive_results(tmp_dict, self.RECURSION_LEVELS)

    def _normalize_row_weights(self, coinstall_dict):
        # Compute an intermediary dictionary that is a row normalized
        # co-install. That is - each coinstalled guid weight is
        # divided by the sum of the weights for all coinstalled guids
        # on this row.
        tmp_dict = {}
        coinstall_total_weight = sum(coinstall_dict.values())
        for coinstall_guid, coinstall_weight in coinstall_dict.items():
            tmp_dict[coinstall_guid] = coinstall_weight / coinstall_total_weight
        return tmp_dict

    def _recursion_penalty(self, level):
        """ Return a factor to apply to the weight for a guid
        recommendation.
        """
        dampener = 1.0 - (1.0 * (self.RECURSION_LEVELS - level) / self.RECURSION_LEVELS)
        dampener *= dampener
        return dampener

    def _compute_recursive_results(self, row_normalized_coinstall, level):
        if level <= 0:
            return row_normalized_coinstall

        # consolidated_coinstall_dict will capture values
        consolidated_coinstall_dict = {}

        # Add this level's guid weight to the consolidated result
        dampener = self._recursion_penalty(level)
        for (
            recommendation_guid,
            recommendation_guid_weight,
        ) in row_normalized_coinstall.items():
            for guid, guid_weight in row_normalized_coinstall.items():
                weight = consolidated_coinstall_dict.get(guid, 0)
                weight += dampener * guid_weight
                consolidated_coinstall_dict[guid] = weight

        # Add in the next level
        level -= 1
        for guid in consolidated_coinstall_dict.keys():
            next_level_coinstalls = self._cache.get_coinstalls(guid, {})
            if next_level_coinstalls != {}:
                # Normalize the next bunch of suggestions
                next_level_coinstalls = self._normalize_row_weights(
                    next_level_coinstalls
                )

                next_level_results = self._compute_recursive_results(
                    next_level_coinstalls, level
                )
                for (next_level_guid, next_level_weight,) in next_level_results.items():
                    weight = consolidated_coinstall_dict.get(guid, 0)
                    weight += next_level_weight
                    consolidated_coinstall_dict[guid] = weight

        # normalize the final results
        return self._normalize_row_weights(consolidated_coinstall_dict)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Test cases for the TAAR CollaborativeRecommender
"""

import contextlib

import fakeredis
import mock
import numpy

from taar.logs.interfaces import ITAARCache
from taar.recommenders.collaborative_recommender import CollaborativeRecommender
from taar.recommenders.collaborative_recommender import positive_hash
from taar.recommenders.redis_cache import TAARCacheRedis
from .noop_fixtures import (
    noop_taarlocale_dataload,
    noop_taarlite_dataload,
    noop_taarensemble_dataload,
    noop_taarsimilarity_dataload,
)

"""
We need to generate a synthetic list of addons and relative weights
for co-occurance.  It's important to note that the
CollaborativeRecommender model expects that addon IDs are hashed using
the Java hash function.
"""


def noop_other_recommenders(stack):
    stack = noop_taarlocale_dataload(stack)
    stack = noop_taarlite_dataload(stack)
    stack = noop_taarsimilarity_dataload(stack)
    stack = noop_taarensemble_dataload(stack)
    return stack


@contextlib.contextmanager
def mock_install_none_mock_data(ctx):
    """
    Overload the 'real' addon model and mapping URLs responses so that
    we always get 404 errors.
    """
    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None

        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_collaborative_item_matrix", return_value="",
            )
        )
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_collaborative_mapping_data", return_value="",
            )
        )

        stack = noop_other_recommenders(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


@contextlib.contextmanager
def mock_install_mock_data(ctx):
    addon_space = [
        {"id": "addon1.id", "name": "addon1.name", "isWebextension": True},
        {"id": "addon2.id", "name": "addon2.name", "isWebextension": True},
        {"id": "addon3.id", "name": "addon3.name", "isWebextension": True},
        {"id": "addon4.id", "name": "addon4.name", "isWebextension": True},
        {"id": "addon5.id", "name": "addon5.name", "isWebextension": True},
    ]

    fake_addon_matrix = []
    for i, addon in enumerate(addon_space):
        row = {
            "id": positive_hash(addon["id"]),
            "features": [0, 0.2, 0.0, 0.1, 0.15],
        }
        row["features"][i] = 1.0
        fake_addon_matrix.append(row)

    fake_mapping = {}
    for addon in addon_space:
        java_hash = positive_hash(addon["id"])
        fake_mapping[str(java_hash)] = addon

    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "_fetch_collaborative_item_matrix",
                return_value=fake_addon_matrix,
            )
        )
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "_fetch_collaborative_mapping_data",
                return_value=fake_mapping,
            )
        )

        stack = noop_other_recommenders(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


def test_cant_recommend(test_ctx):
    with mock_install_mock_data(test_ctx):
        r = CollaborativeRecommender(test_ctx)

        # Test that we can't recommend if we have not enough client info.
        assert not r.can_recommend({})
        assert not r.can_recommend({"installed_addons": []})


def test_can_recommend(test_ctx):
    with mock_install_mock_data(test_ctx):
        r = CollaborativeRecommender(test_ctx)

        # Check that we can recommend if the user has at least an addon.
        assert r.can_recommend(
            {
                "installed_addons": ["uBlock0@raymondhill.net"],
                "client_id": "test-client",
            }
        )


def test_can_recommend_no_model(test_ctx):
    with mock_install_none_mock_data(test_ctx):
        r = CollaborativeRecommender(test_ctx)

        # We should never be able to recommend if something went wrong with the model.
        assert not r.can_recommend({})
        assert not r.can_recommend({"installed_addons": []})
        assert not r.can_recommend({"installed_addons": ["uBlock0@raymondhill.net"]})


def test_empty_recommendations(test_ctx):
    # Tests that the empty recommender always recommends an empty list
    # of addons if we have no addons
    with mock_install_none_mock_data(test_ctx):
        r = CollaborativeRecommender(test_ctx)
        assert not r.can_recommend({})

        # Note that calling recommend() if can_recommend has failed is not
        # defined.


def test_best_recommendation(test_ctx):
    # Make sure the structure of the recommendations is correct and that we
    # recommended the the right addon.
    with mock_install_mock_data(test_ctx):
        r = CollaborativeRecommender(test_ctx)

        # An non-empty set of addons should give a list of recommendations
        fixture_client_data = {
            "installed_addons": ["addon4.id"],
            "client_id": "test_client",
        }
        assert r.can_recommend(fixture_client_data)
        recommendations = r.recommend(fixture_client_data, 1)

        assert isinstance(recommendations, list)
        assert len(recommendations) == 1

        # Verify that addon2 - the most heavy weighted addon was
        # recommended
        result = recommendations[0]
        assert type(result) is tuple
        assert len(result) == 2
        assert result[0] == "addon2.id"
        assert type(result[1]) is numpy.float64
        assert numpy.isclose(result[1], numpy.float64("0.3225"))


def test_recommendation_weights(test_ctx):
    """
    Weights should be ordered greatest to lowest
    """
    with mock_install_mock_data(test_ctx):
        r = CollaborativeRecommender(test_ctx)

        # An non-empty set of addons should give a list of recommendations
        fixture_client_data = {
            "installed_addons": ["addon4.id"],
            "client_id": "test_client",
        }
        assert r.can_recommend(fixture_client_data)
        recommendations = r.recommend(fixture_client_data, 2)
        assert isinstance(recommendations, list)
        assert len(recommendations) == 2

        # Verify that addon2 - the most heavy weighted addon was
        # recommended
        result = recommendations[0]
        assert type(result) is tuple
        assert len(result) == 2
        assert result[0] == "addon2.id"
        assert type(result[1]) is numpy.float64
        assert numpy.isclose(result[1], numpy.float64("0.3225"))

        # Verify that addon2 - the most heavy weighted addon was
        # recommended
        result = recommendations[1]
        assert type(result) is tuple
        assert len(result) == 2
        assert result[0] == "addon5.id"
        assert type(result[1]) is numpy.float64
        assert numpy.isclose(result[1], numpy.float64("0.29"))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.logs.interfaces import IMozLogging
from google.cloud import bigtable
from google.cloud.bigtable import column_family
from google.cloud.bigtable import row_filters
import json
import zlib
import datetime
from taar.settings import (
    BIGTABLE_PROJECT_ID,
    BIGTABLE_INSTANCE_ID,
    BIGTABLE_TABLE_ID,
)
import markus


metrics = markus.get_metrics("taar")


class BigTableProfileController:
    """
    This class implements the profile database in BigTable
    """

    def __init__(self, ctx, project_id, instance_id, table_id):
        self._ctx = ctx
        self._project_id = project_id
        self._instance_id = instance_id
        self._table_id = table_id
        self._column_family_id = "profile"
        self._column_name = "payload".encode()

        # Define the GC policy to retain only the most recent version
        max_age_rule = column_family.MaxAgeGCRule(datetime.timedelta(days=90))
        max_versions_rule = column_family.MaxVersionsGCRule(1)
        self._gc_rule = column_family.GCRuleUnion(
            rules=[max_age_rule, max_versions_rule]
        )

        self._client = bigtable.Client(project=project_id, admin=False)
        self._instance = self._client.instance(self._instance_id)

    def create_table(self):
        # admin needs to be set to True here so that we can create the
        # table
        admin_client = bigtable.Client(project=self._project_id, admin=True)
        instance = admin_client.instance(self._instance_id)
        print("Creating the {} table.".format(self._table_id))
        table = instance.table(self._table_id)

        column_families = {self._column_family_id: self._gc_rule}
        table.create(column_families=column_families)

    def set_client_profile(self, client_profile):

        # Keys must be UTF8 encoded
        row_key = client_profile["client_id"].encode("utf8")

        table = self._instance.table(self._table_id)

        row = table.direct_row(row_key)
        row.set_cell(
            self._column_family_id,
            self._column_name,
            zlib.compress(json.dumps(client_profile).encode("utf8")),
            timestamp=datetime.datetime.utcnow(),
        )
        table.mutate_rows([row])

    def get_client_profile(self, client_id):
        """This fetches a single client record out of GCP BigTable
        """
        try:
            row_key = client_id.encode()
            table = self._instance.table(self._table_id)
            row_filter = row_filters.CellsColumnLimitFilter(1)
            row = table.read_row(row_key, row_filter)
            cell = row.cells[self._column_family_id][self._column_name][0]
            jdata = json.loads(zlib.decompress(cell.value).decode("utf-8"))
            return jdata
        except Exception:
            logger = self._ctx[IMozLogging].get_logger("taar")
            logger.warning(f"Error loading client profile for {client_id}")
            return None


class ProfileFetcher:
    """ Fetch the latest information for a client on the backing
    datastore
    """

    def __init__(self, ctx):
        self._ctx = ctx
        self.logger = self._ctx[IMozLogging].get_logger("taar")
        self.__client = None

    @property
    def _client(self):
        if self.__client is None:
            self.__client = BigTableProfileController(
                self._ctx,
                project_id=BIGTABLE_PROJECT_ID,
                instance_id=BIGTABLE_INSTANCE_ID,
                table_id=BIGTABLE_TABLE_ID,
            )
        return self.__client

    def set_client(self, client):
        self.__client = client

    @metrics.timer_decorator("bigtable_read")
    def get(self, client_id):
        try:
            profile_data = self._client.get_client_profile(client_id)

            if profile_data is None:
                self.logger.debug(
                    "Client profile not found", extra={"client_id": client_id}
                )
                return None

            addon_ids = [
                addon["addon_id"]
                for addon in profile_data.get("active_addons", [])
                if not addon.get("is_system", False)
            ]

            return {
                "client_id": client_id,
                "geo_city": profile_data.get("city", ""),
                "subsession_length": profile_data.get("subsession_length", 0),
                "locale": profile_data.get("locale", ""),
                "os": profile_data.get("os", ""),
                "installed_addons": addon_ids,
                "disabled_addons_ids": profile_data.get("disabled_addons_ids", []),
                "bookmark_count": profile_data.get("places_bookmarks_count", 0),
                "tab_open_count": profile_data.get(
                    "scalar_parent_browser_engagement_tab_open_event_count", 0
                ),
                "total_uri": profile_data.get(
                    "scalar_parent_browser_engagement_total_uri_count", 0
                ),
                "unique_tlds": profile_data.get(
                    "scalar_parent_browser_engagement_unique_domains_count", 0
                ),
            }
        except Exception as e:
            # We just want to catch any kind of error here to make
            # sure nothing breaks
            self.logger.error("Error loading client data", e)
            return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


class MockRecommender:
    """The MockRecommender takes in a map of GUID->weight."""

    def __init__(self, guid_map):
        self._guid_map = guid_map

    def can_recommend(self, *args, **kwargs):
        return True

    def recommend(self, *args, **kwargs):
        return sorted(self._guid_map.items(), key=lambda item: -item[1])


class MockRecommenderFactory:
    """
    A RecommenderFactory provides support to create recommenders.

    The existence of a factory enables injection of dependencies into
    the RecommendationManager and eases the implementation of test
    harnesses.
    """

    def __init__(self, **kwargs):
        mock_legacy = MockRecommender({"abc": 1.0, "bcd": 1.1, "cde": 1.2})
        mock_locale = MockRecommender({"def": 2.0, "efg": 2.1, "fgh": 2.2, "abc": 2.3})
        mock_collaborative = MockRecommender(
            {"ghi": 3.0, "hij": 3.1, "ijk": 3.2, "def": 3.3}
        )
        mock_similarity = MockRecommender(
            {"jkl": 4.0, "klm": 4.1, "lmn": 4.2, "ghi": 4.3}
        )

        self._recommender_factory_map = {
            "legacy": lambda: mock_legacy,
            "collaborative": lambda: mock_collaborative,
            "similarity": lambda: mock_similarity,
            "locale": lambda: mock_locale,
        }

        # Clobber any kwarg passed in recommenders
        for key in self._recommender_factory_map.keys():
            self._recommender_factory_map[key] = kwargs.get(
                key, self._recommender_factory_map[key]
            )

    def get_names(self):
        return self._recommender_factory_map.keys()

    def create(self, recommender_name):
        return self._recommender_factory_map[recommender_name]()


class MockProfileController:
    def __init__(self, mock_profile):
        self._profile = mock_profile

    def get_client_profile(self, client_id):
        return self._profile



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import markus

from taar.logs.interfaces import IMozLogging, ITAARCache
from taar.recommenders.debug import log_timer_debug
from taar.recommenders.ensemble_recommender import (
    EnsembleRecommender,
    is_test_client,
)
from taar.recommenders.randomizer import reorder_guids

metrics = markus.get_metrics("taar")


class RecommenderFactory:
    """
    A RecommenderFactory provides support to create recommenders.

    The existence of a factory enables injection of dependencies into
    the RecommendationManager and eases the implementation of test
    harnesses.
    """

    def __init__(self, ctx):
        self._ctx = ctx
        # This map is set in the default context
        self._recommender_factory_map = self._ctx["recommender_factory_map"]

    def get_names(self):
        return self._recommender_factory_map.keys()

    def create(self, recommender_name):
        return self._recommender_factory_map[recommender_name]()


class RecommendationManager:
    """This class determines which of the set of recommendation
    engines will actually be used to generate recommendations."""

    def __init__(self, ctx):
        """Initialize the user profile fetcher and the recommenders.
        """
        self._ctx = ctx
        self.logger = self._ctx[IMozLogging].get_logger("taar") if self._ctx[IMozLogging] else None

        assert "profile_fetcher" in self._ctx

        self.profile_fetcher = ctx["profile_fetcher"]

        self._ensemble_recommender = EnsembleRecommender(self._ctx.child())

        # The whitelist data is only used for test client IDs

        self._cache = self._ctx[ITAARCache]

    @metrics.timer_decorator("profile_recommendation")
    def recommend(self, client_id, limit, extra_data={}):
        """Return recommendations for the given client.

        The recommendation logic will go through each recommender and
        pick the first one that "can_recommend".

        :param client_id: the client unique id.
        :param limit: the maximum number of recommendations to return.
        :param extra_data: a dictionary with extra client data.
        """

        with log_timer_debug("recommmend executed", self.logger):
            # Read everything from redis now
            with log_timer_debug("redis read", self.logger):
                extra_data["cache"] = self._cache.cache_context()

            if is_test_client(client_id):
                # Just create a stub client_info blob
                client_info = {
                    "client_id": client_id,
                }
            else:
                with log_timer_debug("bigtable fetched data", self.logger):
                    client_info = self.profile_fetcher.get(client_id)

                if client_info is None:
                    self.logger.warning(
                        "Defaulting to empty results.  No client info fetched from storage backend."
                    )
                    return []

            # Fetch back all possible whitelisted addons for this
            # client
            extra_data["guid_randomization"] = True
            whitelist = extra_data["cache"]["whitelist"]
            results = self._ensemble_recommender.recommend(
                client_info, len(whitelist), extra_data
            )

            results = reorder_guids(results, limit)

            self.logger.info(
                f"Client recommendations results",
                extra={'client_id': client_id, 'recs': [r[0] for r in results]}
            )

            return results



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import os
import threading
import redis

from taar.recommenders.cache import TAARCache, RANKING_PREFIX, COINSTALL_PREFIX
from taar.settings import (
    REDIS_HOST,
    REDIS_PORT,
    TAARLITE_MUTEX_TTL,
)

# This marks which of the redis databases is currently
# active for read
ACTIVE_DB = "active_db"

# This is a mutex to block multiple writers from redis
UPDATE_CHECK = "update_mutex|"


class PrefixStripper:
    def __init__(self, prefix, iterator, cast_to_str=False):
        self._prefix = prefix
        self._iter = iterator
        self._cast_to_str = cast_to_str

    def __iter__(self):
        return self

    def __next__(self):
        result = self._iter.__next__()
        result = result[len(self._prefix):]
        if self._cast_to_str:
            result = str(result)
        return result


class TAARCacheRedis(TAARCache):
    """
    This class manages a redis instance to hold onto the taar-lite
    GUID->GUID co-installation data
    """

    _instance = None

    @classmethod
    def get_instance(cls, ctx):
        if cls._instance is None:
            cls._instance = TAARCacheRedis(ctx, i_didnt_read_the_docs=False)
        return cls._instance

    def __init__(self, ctx, i_didnt_read_the_docs=True):
        super(TAARCacheRedis, self).__init__(ctx, i_didnt_read_the_docs)

        self._last_db = None
        # Keep an integer handle (or None) on the last known database
        self._last_db = None

        rcon = self.init_redis_connections()

        self._r0 = rcon[0]
        self._r1 = rcon[1]
        self._r2 = rcon[2]

    def reset(self):
        # Clear out the r0 bookkeeping to reset the database
        return self._r0.flushdb()

    def info(self):
        """
        Dump bookkeeping metadata to logs
        """
        meta = {}
        for key in self._r0.scan_iter():
            meta[key.decode("utf8")] = self._r0.get(key).decode("utf8")
        if len(meta) == 0:
            self.logger.info("Bookkeeping data for TAARLite cache was empty")
        else:
            self.logger.info("TAARLite cache info", extra=meta)

    def init_redis_connections(self):
        """
        Bind connections to redis databases.  This sits in its own
        method to enable mocking for tests
        """
        return {
            0: redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=0),
            1: redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=1),
            2: redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=2),
        }

    def safe_load_data(self):
        """
        This is a multiprocess, multithread safe method to safely load
        data into the cache.

        If a concurrent calls to this method are invoked, only the first
        call will have any effect.
        """
        # Pin the first thread ID to try to update data
        # Note that nx flag so that this is only set if the
        # UPDATE_CHECK is not previously set
        #
        # The thread barrier will autoexpire in 10 minutes in the
        # event of process termination inside the critical section.
        self._r0.set(UPDATE_CHECK, self._ident, nx=True, ex=TAARLITE_MUTEX_TTL)
        self.logger.info(f"UPDATE_CHECK field is set: {self._ident}")

        # This is a concurrency barrier to make sure only the pinned
        # thread can update redis
        update_ident = self._r0.get(UPDATE_CHECK).decode("utf8")
        if update_ident != self._ident:
            self.logger.info(
                "Cache update lock has already been acquired by another process"
            )
            return

        # We're past the thread barrier - load the data and clear the
        # barrier when done
        try:
            self._load_data()
        finally:
            self._r0.delete(UPDATE_CHECK)
            self.logger.info("UPDATE_CHECK field is cleared")

    def _db_get(self, key, default=None, db=None):
        tmp = (db or self._db()).get(key)
        if tmp:
            return json.loads(tmp.decode("utf8"))
        return default

    def _db_set(self, key, val, db):
        db.set(key, json.dumps(val))

    def key_iter_ranking(self):
        return PrefixStripper(
            RANKING_PREFIX, self._db().scan_iter(match=RANKING_PREFIX + "*")
        )

    def key_iter_coinstall(self):
        return PrefixStripper(
            COINSTALL_PREFIX, self._db().scan_iter(match=COINSTALL_PREFIX + "*")
        )

    def is_active(self):
        """
        return True if data is loaded
        """
        # Any value in ACTIVE_DB indicates that data is live
        return self._r0.get(ACTIVE_DB) is not None

    def ensure_db_loaded(self):
        _ = self._db()  # make sure we've computed data from the live redis instance

    def _db(self):
        """
        This dereferences the ACTIVE_DB pointer to get the current
        active redis instance
        """
        active_db = self._r0.get(ACTIVE_DB)

        if active_db is not None:
            db = int(active_db.decode("utf8"))

            if db == 1:
                # Run all callback functions to preprocess model data
                live_db = self._r1
            elif db == 2:
                live_db = self._r2

            self._update_data_callback(db, live_db)

            return live_db

    def _update_data_callback(self, db_num, db):
        """
        Preprocess data when the current redis instance does not match
        the last known instance.
        """
        if db_num == self._last_db:
            return

        self._last_db = db_num

        self._build_cache_context(db)
        self.logger.info("Completed precomputing normalized data")

    @property
    def _ident(self):
        """ pid/thread identity """
        return f"{os.getpid()}_{threading.get_ident()}"

    def _load_data(self):
        active_db = self._r0.get(ACTIVE_DB)
        if active_db is not None:
            active_db = int(active_db.decode("utf8"))
            if active_db == 1:
                next_active_db = 2
            else:
                next_active_db = 1
        else:
            next_active_db = 1

        if next_active_db == 1:
            db = self._r1
        else:
            db = self._r2

        # Clear this database before we do anything with it
        db.flushdb()

        self._copy_data(db)

        self._r0.set(ACTIVE_DB, next_active_db)
        self.logger.info(f"Active DB is set to {next_active_db}")



import re

RE_PLATFORM = re.compile("(linux|windows|macintosh|android|fxios).*firefox")

LINUX = 1
WINDOWS = 2
MACINTOSH = 3
ANDROID = 4
FXIOS = 5

OSNAME_TO_ID = {
    "linux": LINUX,
    "windows": WINDOWS,
    "macintosh": MACINTOSH,
    "android": ANDROID,
    "fxios": FXIOS,
}


def parse_ua(user_agent):
    """
    Return one of the constants for platform selection, otherwise
    return None if the platform cannot be determined.  Any non-firefox
    agent us automatically short circuited to be None.
    """
    ua = user_agent.lower()
    matches = RE_PLATFORM.findall(ua)
    if len(matches) != 1:
        return None
    return OSNAME_TO_ID[matches[0]]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# copy paste from https://github.com/mozilla/srgutil to get rid of this heavy legacy dependency
from taar.logs.interfaces import IMozLogging
import logging.config
import sys


class ContextFilter(logging.Filter):
    """Enhances log messages with contextual information"""

    def __init__(self, logger, func):

        super().__init__()
        self._logger = logger
        self._func = func

    def filter(self, log_record):
        try:
            if self._func:
                self._func(log_record)
        except RuntimeError:
            pass

        return True

    def __enter__(self):
        self._logger.addFilter(self)

    def __exit__(self, type, value, traceback):
        self._logger.removeFilter(self)


class Logging(IMozLogging):
    LOG_NAME = 'srg'
    _log_config = {
        # Note that the formatters.json.logger_name must match
        # loggers.<logger_name> key
        'version': 1,
        'formatters': {
            'json': {
                '()': 'dockerflow.logging.JsonLogFormatter',
                'logger_name': LOG_NAME
            }
        },
        'handlers': {
            'console': {
                'level': 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'json',
                'stream': sys.stdout
            },
        },

        'loggers': {
            LOG_NAME: {
                'handlers': ['console'],
                'level': 'INFO'
            },
        }
    }

    def __init__(self, ctx):
        self._ctx = ctx
        self._logger_prefix = ''
        self._apply_config()

    def set_log_level(self, log_level):
        self._log_config['loggers'][self.LOG_NAME]['level'] = log_level
        self._log_config['handlers']['console']['level'] = log_level
        self._apply_config()

    def _apply_config(self):
        self._logger_prefix = self._log_config['formatters']['json']['logger_name']
        logging.config.dictConfig(self._log_config)

    def get_logger(self, name):
        return logging.getLogger("%s.%s" % (self._logger_prefix, name))



#!/usr/bin/env python
"""A small script that can act as a trust root for installing pip 8

Embed this in your project, and your VCS checkout is all you have to trust. In
a post-peep era, this lets you claw your way to a hash-checking version of pip,
with which you can install the rest of your dependencies safely. All it assumes
is Python 2.7 or better and *some* version of pip already installed. If
anything goes wrong, it will exit with a non-zero status code.

"""
# This is here so embedded copies are MIT-compliant:
# Copyright (c) 2016 Erik Rose
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
from __future__ import print_function
from hashlib import sha256
from os.path import join
from pipes import quote
from shutil import rmtree
from subprocess import check_output
from sys import exit
from tempfile import mkdtemp

try:
    from urllib2 import build_opener, HTTPHandler, HTTPSHandler
except ImportError:
    from urllib.request import build_opener, HTTPHandler, HTTPSHandler
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse  # 3.4


PACKAGES = [
    # Pip has no dependencies, as it vendors everything:
    (
        "https://pypi.python.org/packages/source/p/pip/pip-8.0.2.tar.gz",
        "46f4bd0d8dfd51125a554568d646fe4200a3c2c6c36b9f2d06d2212148439521",
    ),
    # This version of setuptools has only optional dependencies:
    (
        "https://pypi.python.org/packages/source/s/setuptools/"
        "setuptools-19.4.tar.gz",
        "214bf29933f47cf25e6faa569f710731728a07a19cae91ea64f826051f68a8cf",
    ),
    # We require Python 2.7 or later because we don't support wheel's
    # conditional dep on argparse. This version of wheel has no other
    # dependencies:
    (
        "https://pypi.python.org/packages/source/w/wheel/wheel-0.26.0.tar.gz",
        "eaad353805c180a47545a256e6508835b65a8e830ba1093ed8162f19a50a530c",
    ),
]


class HashError(Exception):
    def __str__(self):
        url, path, actual, expected = self.args
        return (
            "{url} did not match the expected hash {expected}. Instead, "
            "it was {actual}. The file (left at {path}) may have been "
            "tampered with.".format(**locals())
        )


def hashed_download(url, temp, digest):
    """Download ``url`` to ``temp``, make sure it has the SHA-256 ``digest``,
    and return its path."""
    # Based on pip 1.4.1's URLOpener but with cert verification removed
    def opener():
        opener = build_opener(HTTPSHandler())
        # Strip out HTTPHandler to prevent MITM spoof:
        for handler in opener.handlers:
            if isinstance(handler, HTTPHandler):
                opener.handlers.remove(handler)
        return opener

    def read_chunks(response, chunk_size):
        while True:
            chunk = response.read(chunk_size)
            if not chunk:
                break
            yield chunk

    response = opener().open(url)
    path = join(temp, urlparse(url).path.split("/")[-1])
    actual_hash = sha256()
    with open(path, "wb") as file:
        for chunk in read_chunks(response, 4096):
            file.write(chunk)
            actual_hash.update(chunk)

    actual_digest = actual_hash.hexdigest()
    if actual_digest != digest:
        raise HashError(url, path, actual_digest, digest)
    return path


def main():
    temp = mkdtemp(prefix="pipstrap-")
    try:
        downloads = [hashed_download(url, temp, digest) for url, digest in PACKAGES]
        check_output(
            "pip install --no-index --no-deps -U "
            + " ".join(quote(d) for d in downloads),
            shell=True,
        )
    except HashError as exc:
        print(exc)
    except Exception:
        rmtree(temp)
        raise
    else:
        rmtree(temp)
        return 0
    return 1


if __name__ == "__main__":
    exit(main())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from taar.profile_fetcher import ProfileFetcher
from taar.profile_fetcher import BigTableProfileController
from google.cloud import bigtable
import copy
import json
import zlib
from mock import MagicMock


class MockProfileController:
    def __init__(self, mock_profile):
        self._profile = mock_profile

    def get_client_profile(self, client_id):
        return self._profile


def test_profile_fetcher_returns_none(test_ctx):
    fetcher = ProfileFetcher(test_ctx)
    fetcher.set_client(MockProfileController(None))
    assert fetcher.get("random-client-id") is None


MOCK_DATA = {
    "profile": {
        u"scalar_parent_browser_engagement_total_uri_count": 791,
        u"city": u"Rome",
        u"scalar_parent_browser_engagement_tab_open_event_count": 46,
        u"subsession_start_date": u"2017-09-20T10:00:00.0+02:00",
        u"subsession_length": 3785,
        u"places_bookmarks_count": 0,
        u"scalar_parent_browser_engagement_unique_domains_count": 11,
        u"os": u"Windows_NT",
        u"active_addons": [
            {u"addon_id": u"e10srollout@mozilla.org"},
            {u"addon_id": u"firefox@getpocket.com"},
            {u"addon_id": u"webcompat@mozilla.org", "is_system": True},
        ],
        u"locale": "it-IT",
    },
    "expected_result": {
        "client_id": "random-client-id",
        "bookmark_count": 0,
        "disabled_addons_ids": [],
        "geo_city": "Rome",
        "os": "Windows_NT",
        "subsession_length": 3785,
        "tab_open_count": 46,
        "total_uri": 791,
        "unique_tlds": 11,
        "installed_addons": ["e10srollout@mozilla.org", "firefox@getpocket.com",],
        "locale": "it-IT",
    },
}


def test_profile_fetcher_returns_dict(test_ctx):
    fetcher = ProfileFetcher(test_ctx)

    mock_data = MOCK_DATA["profile"]
    mock_profile_controller = MockProfileController(mock_data)
    fetcher.set_client(mock_profile_controller)

    # Note that active_addons in the raw JSON source is remapped to
    # 'installed_addons'
    assert fetcher.get("random-client-id") == MOCK_DATA["expected_result"]


def test_dont_crash_without_active_addons(test_ctx):
    mock_data = copy.deepcopy(MOCK_DATA["profile"])
    del mock_data["active_addons"]
    mock_profile_controller = MockProfileController(mock_data)

    fetcher = ProfileFetcher(test_ctx)
    fetcher.set_client(mock_profile_controller)

    expected = copy.deepcopy(MOCK_DATA["expected_result"])
    expected["installed_addons"][:] = []
    assert fetcher.get("random-client-id") == expected


def test_crashy_profile_controller(test_ctx, monkeypatch):
    def mock_bigtable_client(*args, **kwargs):
        class MockClient:
            def __init__(self, *args, **kwargs):
                pass

            def instance(self, *args, **kwargs):
                return MagicMock()

        return MockClient

    monkeypatch.setattr(bigtable, "Client", mock_bigtable_client)

    pc = BigTableProfileController(
        test_ctx, "mock_project_id", "mock_instance_id", "mock_table_id"
    )
    assert pc.get_client_profile("exception_raising_client_id") is None


def test_profile_controller(test_ctx, monkeypatch):
    class MockCell:
        client_profile = {"key": "with_some_data"}
        value = zlib.compress(json.dumps(client_profile).encode("utf8"))

    mc = MockCell()

    def mock_bigtable_client(*args, **kwargs):
        class MockTable:
            def __init__(self, table_id):
                pass

            def read_row(self, *args, **kwargs):
                class MockRow:
                    @property
                    def cells(self):
                        magic_cn = MagicMock()
                        magic_cn.__getitem__.return_value = mc

                        magic_cf = MagicMock()
                        magic_cf.__getitem__.return_value = magic_cn

                        mm = MagicMock()
                        mm.__getitem__.return_value = magic_cf
                        return mm

                return MockRow()

        class MockInstance:
            def table(self, table_id):
                return MockTable(table_id)

        class MockClient:
            def instance(self, *args, **kwargs):
                return MockInstance(*args, **kwargs)

        return MockClient

    monkeypatch.setattr(bigtable, "Client", mock_bigtable_client)

    pc = BigTableProfileController(
        test_ctx, "mock_project_id", "mock_instance_id", "mock_table_id"
    )
    jdata = pc.get_client_profile("a_mock_client")
    assert jdata == {"key": "with_some_data"}


def test_profile_controller_no_user(test_ctx, monkeypatch):
    def mock_bigtable_client(*args, **kwargs):
        class MockTable:
            def __init__(self, table_id):
                pass

            def read_row(self, *args, **kwargs):
                return None

        class MockInstance:
            def table(self, table_id):
                return MockTable(table_id)

        class MockClient:
            def instance(self, *args, **kwargs):
                return MockInstance(*args, **kwargs)

        return MockClient

    monkeypatch.setattr(bigtable, "Client", mock_bigtable_client)

    pc = BigTableProfileController(
        test_ctx, "mock_project_id", "mock_instance_id", "mock_table_id"
    )
    jdata = pc.get_client_profile("a_mock_client")
    assert jdata is None



"""
Test that we can reorder (GUID, weight) tuples based on random
selection based on probability,
"""

from taar.recommenders.randomizer import reorder_guids
from taar.recommenders.randomizer import in_experiment

import numpy as np
from collections import Counter


def most_frequent(List):
    occurence_count = Counter(List)
    return occurence_count.most_common(1)[0][0]


def test_reorder_guids():
    # These weights are selected carefully so that they are different
    # enough that a randomized selection using the weighted inputs
    # will be stable 'enough' that we should be able to pass tests
    # consistently over a sufficiently large sample

    # Fix the random seed so that we get stable results between test
    # runs
    np.random.seed(seed=42)

    guid_weight_tuples = [
        ("guid0", -0.60),
        ("guid1", -0.30),
        ("guid2", 0.09),
        ("guid3", 0.30),
        ("guid4", 2.5),
    ]

    # Run this 100 times to get the average ordering
    results = []
    limit = 4
    for i in range(100):
        results.append(reorder_guids(guid_weight_tuples, size=limit))

    best_result = []
    for i in range(limit):
        best_result.append(most_frequent([row[i] for row in results])[0])

    assert best_result == ["guid4", "guid3", "guid2", "guid1"]


def test_reorder_guids_size_less_than_limit():
    guid_weight_tuples = [
        ("guid0", -0.60),
        ("guid1", -0.30)]
    limit = 4

    reordered = reorder_guids(guid_weight_tuples, size=limit)

    assert len(reordered) == 2


def test_experimental_branch_guid():
    """
    Test the experimental cutoff selection code.

    The evaluation should be stable for a given probability and
    client_id.
    """
    for i in range(10, 100, 10):
        id = hex(i)[2:]
        cutoff = (i + 9.0) / 100

        total = sum([in_experiment(id, cutoff) for i in range(100)])
        assert total == 100

        total = sum([in_experiment(id, cutoff - 0.1) for i in range(100)])
        assert total == 0


def test_in_experiment_zero_prob():
    """
    Test the edge case when some client IDs go to experiment even with 0 probability.
    """
    assert not in_experiment('0ace1ca2a3519332ab93e76a049fe74091fa8fc9063399caa8545dd23f93de5c', xp_prob=0.0)



from abc import ABCMeta, abstractmethod


class AbstractRecommender:
    """Base class for recommenders.

    Subclasses must implement can_recommend and recommend.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def can_recommend(self, client_data, extra_data={}):
        """Tell whether this recommender can recommend the given client."""
        pass

    @abstractmethod
    def recommend(self, client_data, limit, extra_data={}):
        """Return a list of recommendations for the given client."""
        pass

    def __str__(self):
        return self.__class__.__name__



import numpy as np
import bz2
import io
import json
from google.cloud import storage

# TAARLite configuration
from taar.logs.interfaces import IMozLogging, ITAARCache
from taar.settings import (
    TAARLITE_GUID_COINSTALL_BUCKET,
    TAARLITE_GUID_COINSTALL_KEY,
    TAARLITE_GUID_RANKING_KEY,
    TAARLITE_TRUNCATE,
)

# TAAR configuration
from taar.settings import (
    # Locale
    TAAR_LOCALE_BUCKET,
    TAAR_LOCALE_KEY,
    # Collaborative dta
    TAAR_ADDON_MAPPING_BUCKET,
    TAAR_ADDON_MAPPING_KEY,
    TAAR_ITEM_MATRIX_BUCKET,
    TAAR_ITEM_MATRIX_KEY,
    # Similarity data
    TAAR_SIMILARITY_BUCKET,
    TAAR_SIMILARITY_DONOR_KEY,
    TAAR_SIMILARITY_LRCURVES_KEY,
    # Ensemble data
    TAAR_ENSEMBLE_BUCKET,
    TAAR_ENSEMBLE_KEY,
    # Whitelist data
    TAAR_WHITELIST_BUCKET,
    TAAR_WHITELIST_KEY,
    DISABLE_TAAR_LITE,
    DISABLE_ENSEMBLE
)

# taarlite guid guid coinstallation matrix
COINSTALL_PREFIX = "coinstall|"

# taarlite guid guid coinstallation matrix filtered by
# minimum installation threshholds
FILTERED_COINSTALL_PREFIX = "filtered_coinstall|"

# taarlite ranking data
RANKING_PREFIX = "ranking|"

# taarlite minimum installation threshold
MIN_INSTALLS_PREFIX = "min_installs|"

# taarlite map of guid->(sum of coinstall counts)
NORMDATA_COUNT_MAP_PREFIX = "normdata_count_map_prefix|"

# taarlite number of times a GUID shows up per row
# of coinstallation data.
NORMDATA_ROWCOUNT_PREFIX = "normdata_rowcount_prefix|"

# taarlite row nownormalization data
NORMDATA_GUID_ROW_NORM_PREFIX = "normdata_guid_row_norm_prefix|"

# TAAR: Locale data
LOCALE_DATA = "taar_locale_data|"

# TAAR: collaborative data
COLLAB_MAPPING_DATA = "taar_collab_mapping|"
COLLAB_ITEM_MATRIX = "taar_collab_item_matrix|"

# TAAR: similarity data
SIMILARITY_DONORS = "taar_similarity_donors|"
SIMILARITY_LRCURVES = "taar_similarity_lrcurves|"

# TAAR: similarity preprocessed data
SIMILARITY_NUM_DONORS = "taar_similarity_num_donors|"
SIMILARITY_CONTINUOUS_FEATURES = "taar_similarity_continuous_features|"
SIMILARITY_CATEGORICAL_FEATURES = "taar_similarity_categorical_features|"

# TAAR: ensemble weights

ENSEMBLE_WEIGHTS = "taar_ensemble_weights|"

# TAAR: whitelist data
WHITELIST_DATA = "taar_whitelist_data|"


class TAARCache(ITAARCache):
    _instance = None

    """
    Design of this class is heavily influenced by TAARCacheRedis needs.
    In fact, it was extracted from TAARCacheRedis  to be used in
        EnsembleRecommender weights update Spark job independently from Redis
    """

    def __init__(self, ctx, i_didnt_read_the_docs=False):
        """
        Don't call this directly - use get_instance instace
        """
        if i_didnt_read_the_docs:
            raise RuntimeError(
                "You cannot call this method directly - use get_instance"
            )

        self._dict_db = {}

        self._similarity_num_donors = 0
        self._similarity_continuous_features = None
        self._similarity_categorical_features = None

        self._ctx = ctx
        self._last_db = None

        self.logger = None

        moz_logging = self._ctx.get(IMozLogging)
        self.logger = moz_logging.get_logger("taar") if moz_logging else None

    @classmethod
    def get_instance(cls, ctx):
        if cls._instance is None:
            cls._instance = TAARCache(ctx, i_didnt_read_the_docs=False)
        return cls._instance

    # TAARCacheRedis compatibility

    def safe_load_data(self):
        if len(self._dict_db) == 0:
            self._copy_data(self._dict_db)
            self._build_cache_context(self._dict_db)

    def _db_get(self, key, default=None, db=None):
        self.safe_load_data()
        return (db or self._dict_db).get(key, default)

    def _db_set(self, key, val, db):
        self._dict_db[key] = val

    def is_active(self):
        """
        return True if data is loaded
        """
        return len(self._dict_db) > 0

    def ensure_db_loaded(self):
        self.safe_load_data()

    def cache_context(self):
        self.ensure_db_loaded()
        return self._cache_context

    # Getters

    def guid_maps_count_map(self, guid, default=None):
        return self._db_get(NORMDATA_COUNT_MAP_PREFIX + guid) or default

    def guid_maps_rowcount(self, guid, default=None):
        return self._db_get(NORMDATA_ROWCOUNT_PREFIX + guid) or default

    def guid_maps_row_norm(self, guid, default=None):
        return self._db_get(NORMDATA_GUID_ROW_NORM_PREFIX + guid) or default

    def min_installs(self, db):
        """
        Return the floor minimum installed addons that we will
        consider, or 0 if nothing is currently stored in redis
        """
        result = self._db_get(MIN_INSTALLS_PREFIX, db=db)
        if result is None:
            return 0
        return float(result)

    def get_filtered_coinstall(self, guid, default=None):
        tmp = self._db_get(FILTERED_COINSTALL_PREFIX + guid)
        if tmp:
            raw_dict = tmp
            # This truncates the size of the coinstall list for
            # performance reasons
            return dict(
                sorted(raw_dict.items(), key=lambda x: x[1], reverse=True)[:TAARLITE_TRUNCATE]
            )
        return default

    def get_rankings(self, guid, default=None):
        """
        Return the rankings
        """
        return self._db_get(RANKING_PREFIX + guid) or default

    def has_coinstalls_for(self, guid):
        return self._db_get(COINSTALL_PREFIX + guid) is not None

    def get_coinstalls(self, guid, default=None):
        """
        Return a map of GUID:install count that represents the
        coinstallation map for a particular addon GUID
        """
        return self._db_get(COINSTALL_PREFIX + guid) or default

    def top_addons_per_locale(self):
        """
        Get locale data
        """
        return self._db_get(LOCALE_DATA)

    def collab_raw_item_matrix(self):
        """
        Get the taar collaborative item matrix
        """
        return self._db_get(COLLAB_ITEM_MATRIX)

    def collab_addon_mapping(self):
        """
        Get the taar collaborative addon mappin
        """
        return self._db_get(COLLAB_MAPPING_DATA)

    def similarity_donors(self):
        """
        Get the taar similarity donors
        """
        return self._db_get(SIMILARITY_DONORS)

    def similarity_lrcurves(self):
        """
        Get the taar similarity donors
        """
        return self._db_get(SIMILARITY_LRCURVES)

    def similarity_continuous_features(self):
        """
        precomputed similarity recommender continuous features cache
        """
        self.ensure_db_loaded()
        return self._similarity_continuous_features

    def similarity_categorical_features(self):
        """
        precomputed similarity recommender categorical features cache
        """
        self.ensure_db_loaded()
        return self._similarity_categorical_features

    @property
    def similarity_num_donors(self):
        """
        precomputed similarity recommender categorical features cache
        """
        self.ensure_db_loaded()
        return self._similarity_num_donors

    def ensemble_weights(self):
        return self._db_get(ENSEMBLE_WEIGHTS)

    def whitelist_data(self):
        return self._db_get(WHITELIST_DATA)

    # GCS fetching

    def _load_from_gcs(self, bucket, path):
        """
        Load a JSON object off of a GCS bucket and path.

        If the path ends with '.bz2', decompress the object prior to JSON
        decode.
        """
        try:
            with io.BytesIO() as tmpfile:
                client = storage.Client()
                bucket = client.get_bucket(bucket)
                blob = bucket.blob(path)
                blob.download_to_file(tmpfile)
                tmpfile.seek(0)
                payload = tmpfile.read()

                if path.endswith(".bz2"):
                    payload = bz2.decompress(payload)
                    path = path[:-4]

                if path.endswith(".json"):
                    payload = json.loads(payload.decode("utf8"))

                return payload
        except Exception:
            self.logger.exception(f"Error loading from gcs://{bucket}/{path}")

        return None

    def _fetch_coinstall_data(self):
        return self._load_from_gcs(
            TAARLITE_GUID_COINSTALL_BUCKET, TAARLITE_GUID_COINSTALL_KEY
        )

    def _fetch_ranking_data(self):
        return self._load_from_gcs(TAARLITE_GUID_COINSTALL_BUCKET, TAARLITE_GUID_RANKING_KEY)

    def _fetch_locale_data(self):
        return self._load_from_gcs(TAAR_LOCALE_BUCKET, TAAR_LOCALE_KEY)

    def _fetch_collaborative_mapping_data(self):
        return self._load_from_gcs(TAAR_ADDON_MAPPING_BUCKET, TAAR_ADDON_MAPPING_KEY)

    def _fetch_collaborative_item_matrix(self):
        return self._load_from_gcs(TAAR_ITEM_MATRIX_BUCKET, TAAR_ITEM_MATRIX_KEY)

    def _fetch_similarity_donors(self):
        return self._load_from_gcs(TAAR_SIMILARITY_BUCKET, TAAR_SIMILARITY_DONOR_KEY, )

    def _fetch_similarity_lrcurves(self):
        return self._load_from_gcs(TAAR_SIMILARITY_BUCKET, TAAR_SIMILARITY_LRCURVES_KEY, )

    def _fetch_ensemble_weights(self):
        return self._load_from_gcs(TAAR_ENSEMBLE_BUCKET, TAAR_ENSEMBLE_KEY)

    def _fetch_whitelist(self):
        return self._load_from_gcs(TAAR_WHITELIST_BUCKET, TAAR_WHITELIST_KEY)

    # Data update

    def _build_cache_context(self, db):

        self._build_similarity_features_caches(db)

        """
        Fetch from redis once per request
        """
        tmp = {
            # Similarity stuff
            "lr_curves": self.similarity_lrcurves(),
            "num_donors": self.similarity_num_donors,
            "continuous_features": self.similarity_continuous_features(),
            "categorical_features": self.similarity_categorical_features(),
            "donors_pool": self.similarity_donors(),
            # Collaborative
            "addon_mapping": self.collab_addon_mapping(),
            "raw_item_matrix": self.collab_raw_item_matrix(),
            # Locale
            "top_addons_per_locale": self.top_addons_per_locale(),
            # Ensemble
            "whitelist": self.whitelist_data(),
            "ensemble_weights": self.ensemble_weights(),
        }

        def compute_collab_model(val):
            if val not in (None, ""):
                num_rows = len(val)
                num_cols = len(val[0]["features"])

                model = np.zeros(shape=(num_rows, num_cols))
                for index, row in enumerate(val):
                    model[index, :] = row["features"]
            else:
                model = None
            return model

        tmp["collab_model"] = compute_collab_model(tmp["raw_item_matrix"])
        self._cache_context = tmp

    def _build_similarity_features_caches(self, db):
        """
        This function build two feature cache matrices and sets the
        number of donors (self.similarity_num_donors)

        That's the self.categorical_features and
        self.continuous_features attributes.

        One matrix is for the continuous features and the other is for
        the categorical features. This is needed to speed up the similarity
        recommendation process."""
        from taar.recommenders.similarity_recommender import (
            CONTINUOUS_FEATURES,
            CATEGORICAL_FEATURES,
        )

        donors_pool = self._db_get(SIMILARITY_DONORS, db=db)
        if donors_pool is None:
            return

        self._similarity_num_donors = len(donors_pool)

        # Build a numpy matrix cache for the continuous features.
        continuous_features = np.zeros(
            (self.similarity_num_donors, len(CONTINUOUS_FEATURES))
        )

        for idx, d in enumerate(donors_pool):
            features = [d.get(specified_key) for specified_key in CONTINUOUS_FEATURES]
            continuous_features[idx] = features
        self._similarity_continuous_features = continuous_features

        # Build the cache for categorical features.
        categorical_features = np.zeros(
            (self.similarity_num_donors, len(CATEGORICAL_FEATURES)), dtype="object",
        )
        for idx, d in enumerate(donors_pool):
            features = [d.get(specified_key) for specified_key in CATEGORICAL_FEATURES]
            categorical_features[idx] = np.array([features], dtype="object")

        self._similarity_categorical_features = categorical_features

        self.logger.info("Reconstructed matrices for similarity recommender")

    def _update_whitelist_data(self, db):
        """
        Load the TAAR whitelist data
        """
        tmp = self._fetch_whitelist()
        if tmp:
            self._db_set(WHITELIST_DATA, tmp, db)

    def _update_ensemble_data(self, db):
        """
        Load the TAAR ensemble data
        """
        tmp = self._fetch_ensemble_weights()
        if tmp:
            self._db_set(ENSEMBLE_WEIGHTS, tmp["ensemble_weights"], db)

    def _update_similarity_data(self, db):
        """
        Load the TAAR similarity data
        """
        donors = self._fetch_similarity_donors()
        lrcurves = self._fetch_similarity_lrcurves()

        self._db_set(SIMILARITY_DONORS, donors, db)
        self._db_set(SIMILARITY_LRCURVES, lrcurves, db)

    def _update_collab_data(self, db):
        """
        Load the TAAR collaborative data.  This is two parts: an item
        matrix and a mapping of GUIDs
        """
        # Load the item matrix into redis
        item_matrix = self._fetch_collaborative_item_matrix()
        self._db_set(COLLAB_ITEM_MATRIX, item_matrix, db)

        # Load the taar collaborative mapping data
        mapping_data = self._fetch_collaborative_mapping_data()
        self._db_set(COLLAB_MAPPING_DATA, mapping_data, db)

    def _update_locale_data(self, db):
        """
        Load the TAAR locale data
        """
        data = self._fetch_locale_data()
        result = {}
        for locale, guid_list in data.items():
            result[locale] = sorted(guid_list, key=lambda x: x[1], reverse=True)

        self._db_set(LOCALE_DATA, result, db)

    def _update_coinstall_data(self, db):
        """
        Load the TAAR Lite GUID GUID coinstallation data
        """

        data = self._fetch_coinstall_data()

        items = data.items()
        len_items = len(items)

        guid_count_map = {}
        row_count = {}
        guid_row_norm = {}

        for i, (guid, coinstalls) in enumerate(items):

            tmp = dict(
                [(k, v) for (k, v) in coinstalls.items() if v >= self.min_installs(db)]
            )

            self._db_set(FILTERED_COINSTALL_PREFIX + guid, tmp, db)
            rowsum = sum(coinstalls.values())

            for coinstall_guid, coinstall_count in coinstalls.items():
                # Capture the total number of time a GUID was
                # coinstalled with other guids
                guid_count_map.setdefault(coinstall_guid, 0)
                guid_count_map[coinstall_guid] += coinstall_count

                # Capture the unique number of times a GUID is
                # coinstalled with other guids
                row_count.setdefault(coinstall_guid, 0)
                row_count[coinstall_guid] += 1

                if coinstall_guid not in guid_row_norm:
                    guid_row_norm[coinstall_guid] = []
                guid_row_norm[coinstall_guid].append(1.0 * coinstall_count / rowsum)

            self._db_set(COINSTALL_PREFIX + guid, coinstalls, db)
            if i % 1000 == 0:
                self.logger.info(
                    f"Loaded {i + 1} of {len_items} GUID-GUID coinstall records into redis"
                )

        self.logger.info("guidmaps computed - saving")

        for guid, guid_count in guid_count_map.items():
            self._db_set(NORMDATA_COUNT_MAP_PREFIX + guid, guid_count, db)

        for coinstall_guid, coinstall_count in row_count.items():
            self._db_set(NORMDATA_ROWCOUNT_PREFIX + coinstall_guid, coinstall_count, db)

        for coinstall_guid, norm_val in guid_row_norm.items():
            self._db_set(NORMDATA_GUID_ROW_NORM_PREFIX + coinstall_guid, norm_val, db)

        self.logger.info("finished saving guidmaps")

    def _update_rank_data(self, db):

        data = self._fetch_ranking_data()

        items = data.items()
        len_items = len(items)

        for i, (guid, count) in enumerate(items):
            self._db_set(RANKING_PREFIX + guid, count, db)

            if i % 1000 == 0:
                self.logger.info(f"Loaded {i + 1} of {len_items} GUID ranking into redis")

        min_installs = np.mean(list(data.values())) * 0.05
        self._db_set(MIN_INSTALLS_PREFIX, min_installs, db)

        self.logger.info(f"Updated MIN_INSTALLS: {min_installs}")

    def _copy_data(self, db):

        # Update TAARlite
        # it loads a lot of data which we don't need for Ensemble Spark job
        if not DISABLE_TAAR_LITE:
            self._update_rank_data(db)
            self._update_coinstall_data(db)

        # Update TAAR locale data
        self._update_locale_data(db)

        # Update TAAR collaborative data
        self._update_collab_data(db)

        # Update TAAR similarity data
        self._update_similarity_data(db)

        if not DISABLE_ENSEMBLE:
            # Update TAAR ensemble data
            self._update_ensemble_data(db)

            # Update TAAR ensemble data
            self._update_whitelist_data(db)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# copy paste from https://github.com/mozilla/srgutil to get rid of this heavy legacy dependency

"""
A Context is a customizable namespace.

It works like a regular dictionary, but allows you to set a delegate
explicitly to do attribute lookups.

The primary benefit is that the context has a .child() method which
lets you 'lock' a dictionary and clobber the namespace without
affecting parent contexts.

In practice this makes testing easier and allows us to specialize
configuration information as we pass the context through an object
chain.
"""
from taar.logs.interfaces import IMozLogging, ITAARCache


class InvalidInterface(Exception):
    """Raise this when impl() fails to export an implementation"""
    pass


class Context:
    def __init__(self, delegate=None):
        if delegate is None:
            delegate = {}

        self._local_dict = {}
        self._delegate = delegate

    def __contains__(self, key):
        try:
            self[key]
            return True
        except KeyError:
            return False

    def __getitem__(self, key):
        # This is a little tricky, we want to lookup items in our
        # local namespace before we hit the delegate
        try:
            result = self._local_dict[key]
        except KeyError:
            result = self._delegate[key]
        return result

    def get(self, key, default=None):
        try:
            result = self[key]
        except KeyError:
            result = default
        return result

    def __setitem__(self, key, value):
        self._local_dict[key] = value

    def __delitem__(self, key):
        del self._local_dict[key]

    def wrap(self, ctx):
        ctx_child = ctx.child()
        this_child = self.child()
        this_child._delegate = ctx_child
        return this_child

    def child(self):
        """ In general, you should call this immediately in any
        constructor that receives a context """

        return Context(self)

    def impl(self, iface):
        instance = self._local_dict[iface]
        if not isinstance(instance, iface):
            raise InvalidInterface("Instance [%s] doesn't implement requested interface.")
        return instance


def _default_context(cache_cls=None, logger_cls=None, log_level=None):
    ctx = Context()

    from taar.logs.stubs import LoggingStub
    from taar.recommenders.cache import TAARCache

    logger_cls = logger_cls or LoggingStub
    cache_cls = cache_cls or TAARCache

    logger = logger_cls(ctx)
    if log_level:
        logger.set_log_level(log_level)

    ctx[IMozLogging] = logger
    ctx[ITAARCache] = cache_cls.get_instance(ctx)
    return ctx


# The point of all this is to ensure that Ensemble recommender Spark job works with default context
# with minimal dependencies

def default_context(cache_cls=None, logger_cls=None, log_level=None):
    ctx = _default_context(cache_cls, logger_cls, log_level)

    from taar.recommenders import CollaborativeRecommender
    from taar.recommenders import SimilarityRecommender
    from taar.recommenders import LocaleRecommender

    # Note that the EnsembleRecommender is *not* in this map as it
    # needs to ensure that the recommender_map key is installed in the
    # context
    ctx["recommender_factory_map"] = {
        "collaborative": lambda: CollaborativeRecommender(ctx.child()),
        "similarity": lambda: SimilarityRecommender(ctx.child()),
        "locale": lambda: LocaleRecommender(ctx.child()),
    }

    return ctx
